"""
 Copyright (C) 2000, 2001, 2002, 2003 RiskMap srl

 This file is part of QuantLib, a free-software/open-source library
 for financial quantitative analysts and developers - http://quantlib.org/

 QuantLib is free software: you can redistribute it and/or modify it
 under the terms of the QuantLib license.  You should have received a
 copy of the license along with this program; if not, please email
 <quantlib-dev@lists.sf.net>. The license is also available online at
 <https://www.quantlib.org/license.shtml>.

 This program is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the license for more details.
"""

import math
import operator
import QuantLib as ql
import unittest

flag = None


def raiseFlag():
    global flag
    flag = 1


class MarketElementTest(unittest.TestCase):
    def testObservable(self):
        "Testing observability of market elements"
        global flag
        flag = None
        me = ql.SimpleQuote(0.0)
        obs = ql.Observer(raiseFlag)
        obs.registerWith(me)
        me.setValue(3.14)
        if not flag:
            self.fail("Observer was not notified of market element change")
        flag = None
        obs.unregisterWith(me)
        me.setValue(2.71)
        if flag:
            self.fail("Observer was notified after unregistering")

    def testObservableHandle(self):
        "Testing observability of market element handles"
        global flag
        flag = None
        me1 = ql.SimpleQuote(0.0)
        h = ql.RelinkableQuoteHandle(me1)
        obs = ql.Observer(raiseFlag)
        obs.registerWith(h)
        me1.setValue(3.14)
        if not flag:
            self.fail("Observer was not notified of market element change")
        flag = None
        me2 = ql.SimpleQuote(0.0)
        h.linkTo(me2)
        if not flag:
            self.fail("Observer was not notified of market element change")
        flag = None
        obs.unregisterWith(h)
        me2.setValue(2.71)
        if flag:
            self.fail("Observer was notified after unregistering")

    def testObservableErrors(self):
        class Handle:
            def __init__(self, x):
                self.x = x

            def asObservable(self):
                return 10 / self.x

        obs = ql.Observer(raiseFlag)
        self.assertRaises(TypeError, obs.registerWith, 123)
        self.assertRaises(TypeError, obs.registerWith, obs)
        self.assertRaises(TypeError, obs.registerWith, Handle)
        self.assertRaises(ZeroDivisionError, obs.registerWith, Handle(0))
        self.assertRaises(TypeError, obs.registerWith, Handle(1))

    def test_SimpleQuote(self):
        for value in (100, 2.71, 10**100):
            self.assertAlmostEqual(ql.SimpleQuote(value).value(), value)

    def test_DerivedQuote(self):
        sq = ql.SimpleQuote()
        for fn in (math.sqrt, math.exp, math.sin):
            dq = ql.DerivedQuote(ql.QuoteHandle(sq), fn)
            for value in (12, 23, 34):
                sq.setValue(value)
                self.assertAlmostEqual(dq.value(), fn(value))

    def test_CompositeQuote(self):
        sq1 = ql.SimpleQuote()
        sq2 = ql.SimpleQuote()
        for fn in (operator.add, operator.mul, math.hypot):
            cq = ql.CompositeQuote(ql.QuoteHandle(sq1), ql.QuoteHandle(sq2), fn)
            for value in (12, 23, 34):
                sq1.setValue(value)
                sq2.setValue(value + 1)
                self.assertAlmostEqual(cq.value(), fn(value, value + 1))
                self.assertEqual(cq.value1(), value)
                self.assertEqual(cq.value2(), value + 1)

    def test_MultiCompositeQuote(self):
        for fn in (math.fsum, math.prod, lambda xs: sum(x*x for x in xs)**0.5):
            quotes = []
            handles = []
            for i in range(3):
                quotes.append(ql.SimpleQuote(i + 1))
                handles.append(ql.QuoteHandle(quotes[-1]))
                cq = ql.MultiCompositeQuote(ql.QuoteHandleVector(handles), fn)
                for j in range(i + 1):
                    quotes[j].setValue(j * 10 + 1)
                    self.assertAlmostEqual(cq.value(), fn([q.value() for q in quotes]))
                    self.assertEqual(
                        list(map(cq.inputValue, range(len(quotes)))),
                        [q.value() for q in quotes],
                    )


if __name__ == "__main__":
    print("testing QuantLib", ql.__version__)
    unittest.main(verbosity=2)



import QuantLib as ql
import unittest


class SettingsTest(unittest.TestCase):
    def test_properties(self):
        settings = ql.Settings.instance()
        with ql.SavedSettings():
            for v in (ql.Date(1, 1, 2023), ql.Date(2, 2, 2023)):
                settings.evaluationDate = v
                self.assertEqual(settings.evaluationDate, v)
            for v in (True, False):
                settings.enforcesTodaysHistoricFixings = v
                self.assertEqual(settings.enforcesTodaysHistoricFixings, v)
            for v in (True, False):
                settings.includeReferenceDateEvents = v
                self.assertEqual(settings.includeReferenceDateEvents, v)
            for v in (True, False, None):
                settings.includeTodaysCashFlows = v
                self.assertEqual(settings.includeTodaysCashFlows, v)

    def test_saved_settings(self):
        settings = ql.Settings.instance()
        settings.evaluationDate = ql.Date(1, 1, 2023)
        with ql.SavedSettings():
            settings.evaluationDate = ql.Date(2, 2, 2023)
            self.assertEqual(settings.evaluationDate, ql.Date(2, 2, 2023))
        # Test that SavedSettings restores settings on exit.
        self.assertEqual(settings.evaluationDate, ql.Date(1, 1, 2023))


if __name__ == "__main__":
    print("testing QuantLib", ql.__version__)
    unittest.main(verbosity=2)



# Copyright (C) 2021 Marcin Rybacki

# This file is part of QuantLib, a free-software/open-source library
# for financial quantitative analysts and developers - http://quantlib.org/

# QuantLib is free software: you can redistribute it and/or modify it
# under the terms of the QuantLib license.  You should have received a
# copy of the license along with this program; if not, please email
# <quantlib-dev@lists.sf.net>. The license is also available online at
# <https://www.quantlib.org/license.shtml>.

# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the license for more details.


import unittest
import QuantLib as ql


EPSILON = 1.0e-9

CAL = ql.TARGET()

VALUATION_DATE = CAL.adjust(ql.Date(15, ql.March, 2021))

RATE_AVERAGING_MAP = {
    ql.RateAveraging.Compound: "Compounded",
    ql.RateAveraging.Simple: "Simple",
}


def flat_rate(rate):
    return ql.FlatForward(2, CAL, ql.makeQuoteHandle(rate), ql.Actual365Fixed())


def create_ibor_leg(ibor_idx, start, end, payment_lag=0):
    bdc = ibor_idx.businessDayConvention()
    sch = ql.MakeSchedule(
        effectiveDate=start,
        terminationDate=end,
        tenor=ibor_idx.tenor(),
        calendar=CAL,
        convention=bdc,
        backwards=True,
    )
    return ql.IborLeg(
        [1.0],
        sch,
        ibor_idx,
        paymentDayCounter=ibor_idx.dayCounter(),
        paymentConvention=bdc,
        paymentCalendar=CAL,
        paymentLag=payment_lag,
    )


def create_overnight_leg(overnight_idx, start, end, payment_lag=0):
    sch = ql.MakeSchedule(
        effectiveDate=start,
        terminationDate=end,
        tenor=ql.Period(1, ql.Years),
        calendar=CAL,
        convention=ql.Following,
        backwards=True,
    )
    return ql.OvernightLeg(
        [1.0],
        sch,
        overnight_idx,
        paymentDayCounter=ql.Actual365Fixed(),
        paymentConvention=ql.Following,
        paymentCalendar=CAL,
        paymentLag=payment_lag,
    )


def create_fixed_rate_leg(start, end, payment_lag=0):
    sch = ql.MakeSchedule(
        effectiveDate=start,
        terminationDate=end,
        tenor=ql.Period(1, ql.Years),
        calendar=CAL,
        convention=ql.Following,
        backwards=True,
    )
    return ql.FixedRateLeg(
        sch,
        ql.Actual365Fixed(),
        [1.0],
        [0.005],
        paymentAdjustment=ql.Following,
        paymentCalendar=CAL,
        paymentLag=payment_lag,
    )


class CashFlowsTest(unittest.TestCase):
    def setUp(self):
        self.cash_flows = [
            ql.SimpleCashFlow(1.0e6, ql.Date(22, 6, 2022)),
            ql.SimpleCashFlow(5.0e4, ql.Date(22, 6, 2022)),
        ]

    def test_previous_cash_flow_amount(self):
        """Testing previous cash flows amount"""
        reference_date = ql.Date(28, 6, 2022)
        expected_amount = 1.05e6
        include_settlement_date_flows = False
        actual_amount = ql.CashFlows.previousCashFlowAmount(
            self.cash_flows, include_settlement_date_flows, reference_date
        )
        fail_msg = """ Unable to replicate previous cash flow amount:
                            calculated: {actual}
                            expected: {expected}
                   """.format(
            actual=actual_amount, expected=expected_amount
        )
        self.assertEqual(actual_amount, expected_amount, msg=fail_msg)

    def test_next_cash_flow_amount(self):
        """Testing next cash flows amount"""
        reference_date = ql.Date(21, 6, 2022)
        expected_amount = 1.05e6
        include_settlement_date_flows = False
        actual_amount = ql.CashFlows.nextCashFlowAmount(
            self.cash_flows, include_settlement_date_flows, reference_date
        )
        fail_msg = """ Unable to replicate next cash flow amount:
                            calculated: {actual}
                            expected: {expected}
                   """.format(
            actual=actual_amount, expected=expected_amount
        )
        self.assertEqual(actual_amount, expected_amount, msg=fail_msg)


class IborCouponTest(unittest.TestCase):
    def setUp(self):
        ql.Settings.instance().evaluationDate = VALUATION_DATE
        self.nominal_ts_handle = ql.YieldTermStructureHandle(flat_rate(0.007))
        self.ibor_idx = ql.Euribor6M(self.nominal_ts_handle)

    def test_payment_lag(self):
        """Testing payment lag of an Ibor leg"""
        start = ql.Date(17, ql.March, 2021)
        end = ql.Date(17, ql.March, 2031)
        pay_lag = 2
        leg_without_lag = create_ibor_leg(self.ibor_idx, start, end)
        leg_with_lag = create_ibor_leg(self.ibor_idx, start, end, pay_lag)
        for c_f_without_lag, c_f_with_lag in zip(leg_without_lag, leg_with_lag):
            actual_payment_date = c_f_with_lag.date()
            expected_payment_date = CAL.advance(
                c_f_without_lag.date(),
                pay_lag,
                ql.Days,
                self.ibor_idx.businessDayConvention(),
            )

            fail_msg = """ Unable to replicate Ibor coupon payment date:
                            calculated: {actual}
                            expected: {expected}
                       """.format(
                actual=actual_payment_date, expected=expected_payment_date
            )
            self.assertEqual(actual_payment_date, expected_payment_date, msg=fail_msg)


class OvernightCouponTest(unittest.TestCase):
    def setUp(self):
        ql.Settings.instance().evaluationDate = VALUATION_DATE
        self.nominal_ts_handle = ql.YieldTermStructureHandle(flat_rate(0.007))
        self.overnight_idx = ql.Eonia(self.nominal_ts_handle)

    def test_payment_lag(self):
        """Testing payment lag of an overnight leg"""
        start = ql.Date(17, ql.March, 2021)
        end = ql.Date(17, ql.March, 2031)
        pay_lag = 2
        leg_without_lag = create_overnight_leg(self.overnight_idx, start, end)
        leg_with_lag = create_overnight_leg(self.overnight_idx, start, end, pay_lag)
        for c_f_without_lag, c_f_with_lag in zip(leg_without_lag, leg_with_lag):
            actual_payment_date = c_f_with_lag.date()
            expected_payment_date = CAL.advance(
                c_f_without_lag.date(), pay_lag, ql.Days, ql.Following
            )

            fail_msg = """ Unable to replicate overnight coupon payment date:
                            calculated: {actual}
                            expected: {expected}
                       """.format(
                actual=actual_payment_date, expected=expected_payment_date
            )
            self.assertEqual(actual_payment_date, expected_payment_date, msg=fail_msg)


class FixedRateCouponTest(unittest.TestCase):
    def setUp(self):
        ql.Settings.instance().evaluationDate = VALUATION_DATE
        self.nominal_ts_handle = ql.YieldTermStructureHandle(flat_rate(0.007))

    def test_payment_lag(self):
        """Testing payment lag of a fixed rate leg"""
        start = ql.Date(17, ql.March, 2021)
        end = ql.Date(17, ql.March, 2031)
        pay_lag = 2
        leg_without_lag = create_fixed_rate_leg(start, end)
        leg_with_lag = create_fixed_rate_leg(start, end, pay_lag)
        for c_f_without_lag, c_f_with_lag in zip(leg_without_lag, leg_with_lag):
            actual_payment_date = c_f_with_lag.date()
            expected_payment_date = CAL.advance(
                c_f_without_lag.date(), pay_lag, ql.Days, ql.Following
            )

            fail_msg = """ Unable to replicate fixed rate coupon payment date:
                            calculated: {actual}
                            expected: {expected}
                       """.format(
                actual=actual_payment_date, expected=expected_payment_date
            )
            self.assertEqual(actual_payment_date, expected_payment_date, msg=fail_msg)


def create_sub_periods_coupon(
    ibor_idx, start, end, averaging_method=ql.RateAveraging.Compound
):
    reset_schedule = ql.MakeSchedule(
        effectiveDate=start,
        terminationDate=end,
        tenor=ibor_idx.tenor(),
        calendar=ibor_idx.fixingCalendar(),
        convention=ibor_idx.businessDayConvention(),
        backwards=True,
    )
    payment_date = reset_schedule[-1]
    cpn = ql.MultipleResetsCoupon(
        payment_date, 1.0, reset_schedule, ibor_idx.fixingDays(), ibor_idx
    )
    if averaging_method == ql.RateAveraging.Compound:
        cpn.setPricer(ql.CompoundingMultipleResetsPricer())
    else:
        cpn.setPricer(ql.AveragingMultipleResetsPricer())
    return cpn


def create_sub_periods_leg(ibor_idx, start, end, cpn_frequency, averaging_method):
    reset_schedule = ql.MakeSchedule(
        effectiveDate=start,
        terminationDate=end,
        tenor=ibor_idx.tenor(),
        calendar=ibor_idx.fixingCalendar(),
        convention=ibor_idx.businessDayConvention(),
        backwards=True,
    )
    resets_per_coupon = ibor_idx.tenor().frequency() // cpn_frequency
    return ql.MultipleResetsLeg(
        reset_schedule,
        ibor_idx,
        resets_per_coupon,
        [1.0],
        averagingMethod=averaging_method,
    )


def sum_leg_payments(leg):
    return sum([cf.amount() for cf in leg])


def compounded_leg_payment(leg):
    compound = 1.0
    for cf in leg:
        floating_cf = ql.as_floating_rate_coupon(cf)
        year_fraction = floating_cf.accrualPeriod()
        fixing = floating_cf.indexFixing()
        compound *= 1.0 + year_fraction * fixing
    return compound - 1.0


def averaged_leg_payment(leg):
    acc = 0.0
    for cf in leg:
        floating_cf = ql.as_floating_rate_coupon(cf)
        year_fraction = floating_cf.accrualPeriod()
        fixing = floating_cf.indexFixing()
        acc += year_fraction * fixing
    return acc


class SubPeriodsCouponTest(unittest.TestCase):
    def setUp(self):
        ql.Settings.instance().evaluationDate = VALUATION_DATE
        self.nominal_ts_handle = ql.YieldTermStructureHandle(flat_rate(0.007))
        self.ibor_idx = ql.Euribor6M(self.nominal_ts_handle)
        self.ibor_idx.addFixing(ql.Date(10, ql.February, 2021), 0.0085)

    def check_single_period_coupon_replication(self, start, end, averaging):
        ibor_leg = create_ibor_leg(self.ibor_idx, start, end)
        sub_periods_cpn = create_sub_periods_coupon(
            self.ibor_idx, start, end, averaging
        )

        actual_payment = sub_periods_cpn.amount()
        expected_payment = sum_leg_payments(ibor_leg)

        fail_msg = """ Unable to replicate single period coupon payment:
                            calculated: {actual}
                            expected: {expected}
                            start: {start}
                            end: {end}
                   """.format(
            actual=actual_payment, expected=expected_payment, start=start, end=end
        )
        self.assertTrue(abs(actual_payment - expected_payment) < EPSILON, msg=fail_msg)

    def check_multiple_compounded_sub_periods_coupon_replication(self, start, end):
        ibor_leg = create_ibor_leg(self.ibor_idx, start, end)
        sub_periods_cpn = create_sub_periods_coupon(
            self.ibor_idx, start, end, ql.RateAveraging.Compound
        )

        actual_payment = sub_periods_cpn.amount()
        expected_payment = compounded_leg_payment(ibor_leg)

        fail_msg = """ Unable to replicate compounded multiple sub-period coupon payment:
                            calculated: {actual}
                            expected: {expected}
                            start: {start}
                            end: {end}
                   """.format(
            actual=actual_payment, expected=expected_payment, start=start, end=end
        )
        self.assertTrue(abs(actual_payment - expected_payment) < EPSILON, msg=fail_msg)

    def check_multiple_averaged_sub_periods_coupon_replication(self, start, end):
        ibor_leg = create_ibor_leg(self.ibor_idx, start, end)
        sub_periods_cpn = create_sub_periods_coupon(
            self.ibor_idx, start, end, ql.RateAveraging.Simple
        )

        actual_payment = sub_periods_cpn.amount()
        expected_payment = averaged_leg_payment(ibor_leg)

        fail_msg = """ Unable to replicate averaged multiple sub-period coupon payment:
                            calculated: {actual}
                            expected: {expected}
                            start: {start}
                            end: {end}
                   """.format(
            actual=actual_payment, expected=expected_payment, start=start, end=end
        )
        self.assertTrue(abs(actual_payment - expected_payment) < EPSILON, msg=fail_msg)

    def check_sub_periods_leg_replication(self, averaging_method):
        start = ql.Date(18, ql.March, 2021)
        end = ql.Date(18, ql.March, 2022)

        sub_periods_cpn = create_sub_periods_coupon(
            self.ibor_idx, start, end, averaging_method
        )
        sub_periods_leg = create_sub_periods_leg(
            self.ibor_idx, start, end, ql.Annual, averaging_method
        )

        actual_payment = sum_leg_payments(sub_periods_leg)
        expected_payment = sub_periods_cpn.amount()

        fail_msg = """ Unable to replicate sub-period leg payments:
                            calculated: {actual}
                            expected: {expected}
                            averaging: {averaging}
                   """.format(
            actual=actual_payment,
            expected=expected_payment,
            averaging=RATE_AVERAGING_MAP[averaging_method],
        )
        self.assertTrue(abs(actual_payment - expected_payment) < EPSILON, msg=fail_msg)

    def test_regular_single_period_forward_starting_coupon(self):
        """Testing regular single period forward starting coupon"""
        start = ql.Date(15, ql.April, 2021)
        end = ql.Date(15, ql.October, 2021)

        self.check_single_period_coupon_replication(start, end, ql.RateAveraging.Simple)
        self.check_single_period_coupon_replication(
            start, end, ql.RateAveraging.Compound
        )

    def test_regular_single_period_coupon_after_fixing(self):
        """Testing regular single period coupon after fixing"""
        start = ql.Date(12, ql.February, 2021)
        end = ql.Date(12, ql.August, 2021)

        self.check_single_period_coupon_replication(start, end, ql.RateAveraging.Simple)
        self.check_single_period_coupon_replication(
            start, end, ql.RateAveraging.Compound
        )

    def test_irregular_single_period_coupon_after_fixing(self):
        """Testing irregular single period coupon after fixing"""
        start = ql.Date(12, ql.February, 2021)
        end = ql.Date(12, ql.June, 2021)

        self.check_single_period_coupon_replication(start, end, ql.RateAveraging.Simple)
        self.check_single_period_coupon_replication(
            start, end, ql.RateAveraging.Compound
        )

    def test_regular_compounded_forward_starting_coupon_with_multiple_sub_periods(self):
        """Testing regular forward starting coupon with multiple compounded sub-periods"""
        start = ql.Date(15, ql.April, 2021)
        end = ql.Date(15, ql.April, 2022)

        self.check_multiple_compounded_sub_periods_coupon_replication(start, end)

    def test_regular_averaged_forward_starting_coupon_with_multiple_sub_periods(self):
        """Testing regular forward starting coupon with multiple averaged sub-periods"""
        start = ql.Date(15, ql.April, 2021)
        end = ql.Date(15, ql.April, 2022)

        self.check_multiple_averaged_sub_periods_coupon_replication(start, end)

    def test_sub_periods_leg_cash_flows(self):
        """Testing sub-periods leg replication"""
        self.check_sub_periods_leg_replication(ql.RateAveraging.Compound)
        self.check_sub_periods_leg_replication(ql.RateAveraging.Simple)

    def test_casting(self):
        """Testing casting to sub periods coupon"""
        start = ql.Date(18, ql.March, 2021)
        end = ql.Date(18, ql.March, 2022)
        sub_periods_leg = create_sub_periods_leg(
            self.ibor_idx, start, end, ql.Annual, ql.RateAveraging.Compound
        )
        cf = sub_periods_leg[0]
        self.assertTrue(not isinstance(cf, ql.MultipleResetsCoupon))
        self.assertTrue(
            isinstance(ql.as_multiple_resets_coupon(cf), ql.MultipleResetsCoupon)
        )

    def test_sub_period_coupon_fixing_dates(self):
        """Testing sub-period coupon fixing dates"""
        start = ql.Date(15, ql.April, 2021)
        end = ql.Date(15, ql.April, 2022)
        cpn = ql.as_multiple_resets_coupon(
            create_sub_periods_coupon(self.ibor_idx, start, end)
        )
        actual_dates = cpn.fixingDates()
        expected_dates = (ql.Date(13, 4, 2021), ql.Date(13, 10, 2021))

        fail_msg = """ Unable to replicate sub-period coupon fixing dates:
                            calculated: {actual}
                            expected: {expected}
                       """.format(
            actual=actual_dates, expected=expected_dates
        )
        self.assertTupleEqual(actual_dates, expected_dates, msg=fail_msg)

    def test_sub_period_coupon_value_dates(self):
        """Testing sub-period coupon value dates"""
        start = ql.Date(15, ql.April, 2021)
        end = ql.Date(15, ql.April, 2022)
        cpn = ql.as_multiple_resets_coupon(
            create_sub_periods_coupon(self.ibor_idx, start, end)
        )
        actual_dates = cpn.valueDates()
        expected_dates = (
            ql.Date(15, 4, 2021),
            ql.Date(15, 10, 2021),
            ql.Date(19, 4, 2022),
        )

        fail_msg = """ Unable to replicate sub-period coupon value dates:
                            calculated: {actual}
                            expected: {expected}
                       """.format(
            actual=actual_dates, expected=expected_dates
        )
        self.assertTupleEqual(actual_dates, expected_dates, msg=fail_msg)

    def test_sub_period_coupon_rate_spread(self):
        """Testing sub-period coupon rate spread"""
        start = ql.Date(15, ql.April, 2021)
        end = ql.Date(15, ql.April, 2022)
        cpn = ql.as_multiple_resets_coupon(
            create_sub_periods_coupon(self.ibor_idx, start, end)
        )
        actual_spread = cpn.rateSpread()
        expected_spread = 0.0

        fail_msg = """ Unable to replicate sub-period coupon rate spread:
                            calculated: {actual}
                            expected: {expected}
                       """.format(
            actual=actual_spread, expected=expected_spread
        )
        self.assertEqual(actual_spread, expected_spread, msg=fail_msg)


if __name__ == "__main__":
    print("testing QuantLib", ql.__version__)
    unittest.main(verbosity=2)



# ---
# jupyter:
#   jupytext:
#     formats: py:light
#     text_representation:
#       extension: .py
#       format_name: light
#       format_version: '1.5'
#       jupytext_version: 1.14.5
#   kernelspec:
#     display_name: Python 3 (ipykernel)
#     language: python
#     name: python3
# ---

# # Callable bonds
#
# This file is part of QuantLib, a free-software/open-source library
# for financial quantitative analysts and developers - https://www.quantlib.org/
#
# QuantLib is free software: you can redistribute it and/or modify it under the
# terms of the QuantLib license.  You should have received a copy of the
# license along with this program; if not, please email
# <quantlib-dev@lists.sf.net>. The license is also available online at
# <https://www.quantlib.org/license.shtml>.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the license for more details.

import QuantLib as ql
import numpy as np

calcDate = ql.Date(16, 8, 2006)
ql.Settings.instance().evaluationDate = calcDate

dayCount = ql.ActualActual(ql.ActualActual.Bond)
rate = 0.0465
termStructure = ql.FlatForward(calcDate, rate, dayCount, ql.Compounded, ql.Semiannual)
term_structure_handle = ql.RelinkableYieldTermStructureHandle(termStructure)

callabilitySchedule = ql.CallabilitySchedule()
callPrice = 100.0
callDate = ql.Date(15, ql.September, 2006);
nc = ql.NullCalendar()

# Number of calldates is 24
for i in range(0, 24):
    callabilityPrice  = ql.BondPrice(callPrice, ql.BondPrice.Clean)
    callabilitySchedule.append(ql.Callability(callabilityPrice, ql.Callability.Call, callDate))
    callDate = nc.advance(callDate, 3, ql.Months)

issueDate = ql.Date(16, ql.September, 2004)
maturityDate = ql.Date(15, ql.September, 2012)
calendar = ql.UnitedStates(ql.UnitedStates.GovernmentBond)
tenor = ql.Period(ql.Quarterly)
accrualConvention = ql.Unadjusted
schedule = ql.Schedule(issueDate, maturityDate, tenor, calendar,
                       accrualConvention, accrualConvention, ql.DateGeneration.Backward, False)

settlement_days = 3
faceAmount = 100
accrual_daycount = ql.ActualActual(ql.ActualActual.Bond)
coupon = 0.025
bond = ql.CallableFixedRateBond(settlement_days, faceAmount, schedule,
                                [coupon], accrual_daycount, ql.Following,
                                faceAmount, issueDate, callabilitySchedule)

def engine(a, s, grid_points):
    model = ql.HullWhite(term_structure_handle, a, s)
    return ql.TreeCallableFixedRateBondEngine(model, grid_points)

# 6% mean reversion and 20% volatility
bond.setPricingEngine(engine(0.06, 0.20, 40))
print("Bond price: ", bond.cleanPrice())

# 3% mean reversion and 15% volatility
bond.setPricingEngine(engine(0.03, 0.15, 40))
print("Bond price: ", bond.cleanPrice())



"""
 Copyright (C) 2020 Marcin Rybacki

 This file is part of QuantLib, a free-software/open-source library
 for financial quantitative analysts and developers - http://quantlib.org/

 QuantLib is free software: you can redistribute it and/or modify it
 under the terms of the QuantLib license.  You should have received a
 copy of the license along with this program; if not, please email
 <quantlib-dev@lists.sf.net>. The license is also available online at
 <https://www.quantlib.org/license.shtml>.

 This program is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the license for more details.
"""

import unittest
import QuantLib as ql


EPSILON = 1.e-8

# Hypothetical market data
EUR_ZERO_RATES = [(ql.Period(1, ql.Days), 0.0048),
                  (ql.Period(1, ql.Years), 0.0048),
                  (ql.Period(2, ql.Years), 0.00475),
                  (ql.Period(3, ql.Years), 0.005),
                  (ql.Period(5, ql.Years), 0.0055),
                  (ql.Period(10, ql.Years), 0.007)]

EUR_BEI_SWAP_RATES = [(ql.Period(1, ql.Years), 0.0301),
                      (ql.Period(2, ql.Years), 0.0299),
                      (ql.Period(3, ql.Years), 0.0305),
                      (ql.Period(5, ql.Years), 0.0315),
                      (ql.Period(10, ql.Years), 0.0355)]

# Source:
# https://ec.europa.eu/eurostat/web/products-datasets/-/teicp240.
EU_FIXING_DATA = [(ql.Date(1, ql.April, 2018), 103.11),
                  (ql.Date(1, ql.May, 2018), 103.64),
                  (ql.Date(1, ql.June, 2018), 103.76),
                  (ql.Date(1, ql.July, 2018), 103.41),
                  (ql.Date(1, ql.August, 2018), 103.58)]

CAL = ql.TARGET()

DAY_COUNTER = ql.ActualActual(ql.ActualActual.ISDA)

BDC = ql.ModifiedFollowing

VALUATION_DATE = CAL.adjust(ql.Date(10, ql.September, 2018))

OBSERVATION_LAG = ql.Period(3, ql.Months)


def create_inflation_swap_helper(
        reference_date,
        inflation_data,
        inflation_index,
        interpolation,
        observation_lag=OBSERVATION_LAG,
        calendar=CAL,
        business_day_convention=BDC,
        day_counter=DAY_COUNTER):
    maturity = CAL.advance(reference_date, inflation_data[0])
    quote = ql.makeQuoteHandle(inflation_data[1])
    return ql.ZeroCouponInflationSwapHelper(
        quote,
        observation_lag,
        maturity,
        calendar,
        business_day_convention,
        day_counter,
        inflation_index,
        interpolation,
    )


def build_nominal_term_structure(
        reference_date,
        nominal_data):
    nominal_dc = ql.Actual365Fixed()
    dates = [CAL.advance(reference_date, x[0]) for x in nominal_data]
    rates = [x[1] for x in nominal_data]
    return ql.ZeroCurve(dates, rates, nominal_dc)


def build_hicp_index(
        fixing_data,
        inflation_crv_handle):
    index = ql.EUHICP(inflation_crv_handle)
    for x in fixing_data:
        # force override in case of multiple use
        index.addFixing(x[0], x[1], True)
    return index


SEASONAL = {ql.January: 1.0, ql.February: 1.01, ql.March: 1.011,
            ql.April: 1.009, ql.May: 1.008, ql.June: 1.012,
            ql.July: 1.0078, ql.August: 1.006,
            ql.September: 1.0085, ql.October: 1.0096,
            ql.November: 1.0067, ql.December: 1.0055}


def construct_seasonality(reference_date):
    frequency = ql.Monthly
    seasonality_base_date = ql.Date(1, ql.January, reference_date.year())
    factors = list(SEASONAL.values())
    return ql.MultiplicativePriceSeasonality(
        seasonality_base_date, frequency, factors)


def build_inflation_term_structure(
        reference_date,
        zero_coupon_swaps_data,
        inflation_index,
        interpolation,
        observation_lag=OBSERVATION_LAG,
        include_seasonality=False):
    helpers = [create_inflation_swap_helper(reference_date,
                                            x,
                                            inflation_index,
                                            interpolation)
               for x in zero_coupon_swaps_data]
    cpi_term_structure = ql.PiecewiseZeroInflation(
        reference_date,
        inflation_index.lastFixingDate(),
        inflation_index.frequency(),
        DAY_COUNTER,
        helpers)
    if include_seasonality:
        seasonality = construct_seasonality(reference_date)
        cpi_term_structure.setSeasonality(seasonality)
    return cpi_term_structure


def create_inflation_swap(
        inflation_idx,
        start_date,
        end_date,
        rate,
        interpolation,
        observation_lag=OBSERVATION_LAG,
        nominal=1.e6,
        payer=ql.Swap.Payer):
    return ql.ZeroCouponInflationSwap(
        payer,
        nominal,
        start_date,
        end_date,
        CAL,
        BDC,
        DAY_COUNTER,
        rate,
        inflation_idx,
        observation_lag,
        interpolation)


def interpolate_historic_index(
        inflation_idx, fixing_date, observation_lag=OBSERVATION_LAG):
    first_dt = ql.Date(1, fixing_date.month(), fixing_date.year())
    second_dt = ql.Date.endOfMonth(fixing_date) + 1
    slope_numerator = fixing_date - first_dt
    slope_denominator = (
        (second_dt + observation_lag) - (first_dt + observation_lag))
    slope = float(slope_numerator) / float(slope_denominator)
    return inflation_idx.fixing(first_dt) + slope * (
        inflation_idx.fixing(second_dt) - inflation_idx.fixing(first_dt))


class InflationTest(unittest.TestCase):
    def setUp(self):
        ql.Settings.instance().evaluationDate = VALUATION_DATE
        self.inflation_ts_handle = ql.RelinkableZeroInflationTermStructureHandle()
        self.nominal_ts_handle = ql.RelinkableYieldTermStructureHandle()
        self.nominal_ts_handle.linkTo(
            build_nominal_term_structure(VALUATION_DATE, EUR_ZERO_RATES))
        self.discount_engine = ql.DiscountingSwapEngine(self.nominal_ts_handle)

    def test_par_swap_pricing_fom_indexation_without_seasonality(self):
        """Testing pricing of par inflation swap for First-Of-Month indexation"""

        inflation_idx = build_hicp_index(
            EU_FIXING_DATA, self.inflation_ts_handle)
        inflation_ts = build_inflation_term_structure(
            VALUATION_DATE,
            EUR_BEI_SWAP_RATES,
            inflation_idx,
            ql.CPI.Flat)
        self.inflation_ts_handle.linkTo(inflation_ts)

        zciis = create_inflation_swap(
            inflation_idx,
            VALUATION_DATE,
            CAL.advance(VALUATION_DATE, ql.Period(10, ql.Years)),
            0.0355,
            ql.CPI.Flat)
        zciis.setPricingEngine(self.discount_engine)
        npv = zciis.NPV()
        # Check whether swap prices to par
        fail_msg = """ Failed to price zero coupon inflation swap to par:
                            index: {inflation_idx}
                            end date: {end_date}
                            observation lag: {observation_lag}
                            npv: {npv}
                            expected npv: {expected_npv}
                            tolerance: {tolerance}
                   """.format(inflation_idx=inflation_idx.familyName(),
                              end_date=zciis.maturityDate(),
                              observation_lag=OBSERVATION_LAG,
                              npv=npv,
                              expected_npv=0.0,
                              tolerance=EPSILON)
        self.assertTrue(
            abs(npv) < EPSILON,
            msg=fail_msg)

    def test_inflation_leg_payment_fom_indexation_without_seasonality(self):
        """Testing inflation leg payment for First-Of-Month indexation"""

        inflation_idx = build_hicp_index(
            EU_FIXING_DATA, self.inflation_ts_handle)
        inflation_ts = build_inflation_term_structure(
            VALUATION_DATE,
            EUR_BEI_SWAP_RATES,
            inflation_idx,
            ql.CPI.Flat)
        self.inflation_ts_handle.linkTo(inflation_ts)

        zciis = create_inflation_swap(
            inflation_idx,
            VALUATION_DATE,
            CAL.advance(VALUATION_DATE, ql.Period(10, ql.Years)),
            0.0355,
            ql.CPI.Flat)
        zciis.setPricingEngine(self.discount_engine)

        inflation_cf = ql.as_indexed_cashflow(
            zciis.inflationLeg()[0])
        # Obtaining base index for the inflation swap
        swap_base_dt = inflation_cf.baseDate()
        swap_base_fixing = inflation_idx.fixing(swap_base_dt)
        # Replicate fixing projection
        fixing_dt = inflation_cf.fixingDate()
        ts_base_dt = inflation_ts.baseDate()
        ts_base_fixing = inflation_idx.fixing(ts_base_dt)
        # Apply FOM indexation rule
        effective_fixing_dt = ql.Date(
            1, fixing_dt.month(), fixing_dt.year())
        fraction = inflation_ts.dayCounter().yearFraction(
            ts_base_dt, effective_fixing_dt)
        t = inflation_ts.timeFromReference(effective_fixing_dt)
        zero_rate = inflation_ts.zeroRate(t)
        expected_fixing = ts_base_fixing * (
            1.0 + zero_rate)**fraction

        expected_inflation_leg_payment = (
            expected_fixing / swap_base_fixing - 1.0) * inflation_cf.notional()
        actual_inflation_leg_payment = inflation_cf.amount()

        fail_msg = """ Failed to replicate inflation leg payment
                       for First-Of-Month indexation:
                            index: {inflation_idx}
                            end date: {end_date}
                            observation lag: {observation_lag}
                            inflation leg payment: {actual_payment}
                            replicated payment: {expected_payment}
                            tolerance: {tolerance}
                   """.format(inflation_idx=inflation_idx.familyName(),
                              end_date=zciis.maturityDate(),
                              observation_lag=OBSERVATION_LAG,
                              actual_payment=actual_inflation_leg_payment,
                              expected_payment=expected_inflation_leg_payment,
                              tolerance=EPSILON)
        self.assertAlmostEqual(
            first=actual_inflation_leg_payment,
            second=expected_inflation_leg_payment,
            delta=EPSILON,
            msg=fail_msg)

    def test_lagged_fixing_method(self):
        """Testing lagged fixing method"""

        inflation_idx = build_hicp_index(
            EU_FIXING_DATA, self.inflation_ts_handle)
        inflation_ts = build_inflation_term_structure(
            VALUATION_DATE,
            EUR_BEI_SWAP_RATES,
            inflation_idx,
            ql.CPI.Flat)
        self.inflation_ts_handle.linkTo(inflation_ts)

        maturity_date = ql.Date(25, ql.October, 2027)
        lag = ql.Period(3, ql.Months)
        indexation = ql.CPI.Flat

        actual_fixing = ql.CPI.laggedFixing(inflation_idx, maturity_date, lag, indexation)
        expected_fixing = inflation_idx.fixing(ql.Date(1, ql.July, 2027))

        fail_msg = """ Failed to replicate lagged fixing:
                            index: {inflation_idx}
                            actual fixing: {actual_fixing}
                            expected fixing: {expected_fixing}
                            tolerance: {tolerance}
                   """.format(inflation_idx=inflation_idx.familyName(),
                              actual_fixing=actual_fixing,
                              expected_fixing=expected_fixing,
                              tolerance=EPSILON)
        self.assertAlmostEqual(
            first=actual_fixing,
            second=expected_fixing,
            msg=fail_msg,
            delta=EPSILON)


if __name__ == '__main__':
    print("testing QuantLib", ql.__version__)
    unittest.main(verbosity=2)



"""
 Copyright (C) 2000, 2001, 2002, 2003 RiskMap srl

 This file is part of QuantLib, a free-software/open-source library
 for financial quantitative analysts and developers - http://quantlib.org/

 QuantLib is free software: you can redistribute it and/or modify it
 under the terms of the QuantLib license.  You should have received a
 copy of the license along with this program; if not, please email
 <quantlib-dev@lists.sf.net>. The license is also available online at
 <https://www.quantlib.org/license.shtml>.

 This program is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the license for more details.
"""

import QuantLib as ql
import unittest
import math


class IntegralTest(unittest.TestCase):
    def Gauss(self, x):
        return math.exp(-x * x / 2.0) / math.sqrt(2 * math.pi)

    def singleTest(self, I):
        tolerance = 1e-4
        cases = [
            ["f(x) = 1", lambda x: 1, 0.0, 1.0, 1.0],
            ["f(x) = x", lambda x: x, 0.0, 1.0, 0.5],
            ["f(x) = x^2", lambda x: x * x, 0.0, 1.0, 1.0 / 3.0],
            ["f(x) = sin(x)", math.sin, 0.0, math.pi, 2.0],
            ["f(x) = cos(x)", math.cos, 0.0, math.pi, 0.0],
            ["f(x) = Gauss(x)", self.Gauss, -10.0, 10.0, 1.0],
        ]

        for tag, f, a, b, expected in cases:
            calculated = I(f, a, b)
            if not (abs(calculated - expected) <= tolerance):
                self.fail(
                    """
integrating %(tag)s
    calculated: %(calculated)f
    expected  : %(expected)f
                      """
                    % locals()
                )

    def testSegment(self):
        "Testing segment integration"
        self.singleTest(ql.SegmentIntegral(10000))

    def testTrapezoid(self):
        "Testing trapezoid integration"
        self.singleTest(ql.TrapezoidIntegralDefault(1.0e-4, 1000))

    def testSimpson(self):
        "Testing Simpson integration"
        self.singleTest(ql.SimpsonIntegral(1.0e-4, 1000))

    def testKronrod(self):
        "Testing Gauss-Kronrod integration"
        self.singleTest(ql.GaussKronrodAdaptive(1.0e-4))


if __name__ == "__main__":
    print("testing QuantLib", ql.__version__)
    unittest.main(verbosity=2)



# ---
# jupyter:
#   jupytext:
#     formats: py:light
#     text_representation:
#       extension: .py
#       format_name: light
#       format_version: '1.5'
#       jupytext_version: 1.11.2
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

# # ISDA CDS engine
#
# This file is part of QuantLib, a free-software/open-source library
# for financial quantitative analysts and developers - https://www.quantlib.org/
#
# QuantLib is free software: you can redistribute it and/or modify it under the
# terms of the QuantLib license.  You should have received a copy of the
# license along with this program; if not, please email
# <quantlib-dev@lists.sf.net>. The license is also available online at
# <https://www.quantlib.org/license.shtml>.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the license for more details.

import QuantLib as ql
import pandas as pd

interactive = 'get_ipython' in globals()

trade_date = ql.Date(21,5,2009)
ql.Settings.instance().evaluationDate = trade_date

ql.IborCoupon.createAtParCoupons()

dep_tenors = [1,2,3,6,9,12]
dep_quotes = [0.003081,0.005525,0.007163,0.012413,0.014,0.015488]
isdaRateHelpers = [ql.DepositRateHelper(dep_quotes[i],
                                        dep_tenors[i]*ql.Period(ql.Monthly),
                                        2,ql.WeekendsOnly(),
                                        ql.ModifiedFollowing,
                                        False,ql.Actual360())
                   for i in range(len(dep_tenors))]

swap_tenors = [2, 3, 4, 5, 6, 7, 8, 9, 10, 12, 15, 20, 25, 30]
swap_quotes = [0.011907,
               0.01699,
               0.021198,
               0.02444,
               0.026937,
               0.028967,
               0.030504,
               0.031719,
               0.03279,
               0.034535,
               0.036217,
               0.036981,
               0.037246,
               0.037605]

isda_ibor = ql.IborIndex('IsdaIbor',3*ql.Period(ql.Monthly),2,
                         ql.USDCurrency(),ql.WeekendsOnly(),
                         ql.ModifiedFollowing,False,ql.Actual360())
isdaRateHelpers = isdaRateHelpers + [
    ql.SwapRateHelper(swap_quotes[i],swap_tenors[i]*ql.Period(ql.Annual),
                      ql.WeekendsOnly(),ql.Semiannual,ql.ModifiedFollowing,
                      ql.Thirty360(ql.Thirty360.BondBasis),isda_ibor)
    for i in range(len(swap_tenors))]

spot_date = ql.WeekendsOnly().advance(trade_date, 2 * ql.Period(ql.Daily))

# Technically, the model requires the discount factor to be 1 at spot;
# but we can't do that and also have the discount curve extend back to
# the trade date.  For the time being, we'll keep discount = 1 at trade.
# The results match anyway.

swap_curve = ql.PiecewiseFlatForward(trade_date, isdaRateHelpers, ql.Actual365Fixed())
discountCurve = ql.YieldTermStructureHandle(swap_curve)

probabilityCurve = ql.RelinkableDefaultProbabilityTermStructureHandle()

termDates = [ql.Date(20, 6, 2010),
             ql.Date(20, 6, 2011),
             ql.Date(20, 6, 2012),
             ql.Date(20, 6, 2016),
             ql.Date(20, 6, 2019)]

spreads = [0.001, 0.1]
recoveries = [0.2, 0.4]

markitValues = [97798.29358, #0.001
                97776.11889, #0.001
                -914971.5977, #0.1
                -894985.6298, #0.1
                186921.3594, #0.001
                186839.8148, #0.001
                -1646623.672, #0.1
                -1579803.626, #0.1
                274298.9203,
                274122.4725,
                -2279730.93,
                -2147972.527,
                592420.2297,
                591571.2294,
                -3993550.206,
                -3545843.418,
                797501.1422,
                795915.9787,
                -4702034.688,
                -4042340.999]

tolerance = 1.0e-2

l = 0
distance = 0

data = []
upfront_date = ql.WeekendsOnly().advance(trade_date, 3 * ql.Period(ql.Daily))
for termDate in termDates:
    for spread in spreads:
        for recovery in recoveries:

            cdsSchedule = ql.Schedule(trade_date, termDate,
                                      3*ql.Period(ql.Monthly),
                                      ql.WeekendsOnly(),
                                      ql.Following, ql.Unadjusted,
                                      ql.DateGeneration.CDS, False)

            quotedTrade = ql.CreditDefaultSwap(
                ql.Protection.Buyer,10000000,0,spread,cdsSchedule,
                ql.Following,ql.Actual360(),True,True,trade_date,
                upfront_date, ql.FaceValueClaim(), ql.Actual360(True))

            h = quotedTrade.impliedHazardRate(0,discountCurve,ql.Actual365Fixed(),
                                              recovery,1e-10,
                                              ql.CreditDefaultSwap.ISDA)

            probabilityCurve.linkTo(
                ql.FlatHazardRate(0,ql.WeekendsOnly(),
                                  ql.makeQuoteHandle(h),
                                  ql.Actual365Fixed()))

            engine = ql.IsdaCdsEngine(probabilityCurve,recovery,discountCurve)
            conventionalTrade = ql.CreditDefaultSwap(
                ql.Protection.Buyer,10000000,0,0.01,cdsSchedule,
                ql.Following,ql.Actual360(),True,True,trade_date,
                upfront_date, ql.FaceValueClaim(), ql.Actual360(True))
            conventionalTrade.setPricingEngine(engine)

            upfront = conventionalTrade.notional() * conventionalTrade.fairUpfront()

            data.append(
                (termDate,
                 spread,
                 recovery,
                 h,
                 upfront,
                 markitValues[l],
                 abs(upfront-markitValues[l]),
                 abs(upfront-markitValues[l])<tolerance)
            )
            distance = distance + abs(upfront-markitValues[l])

            l = l + 1

df = pd.DataFrame(data, columns=["Term date", "Spread", "Recovery",
                                 "Hazard rate", "Upfront", "Markit value", "Distance", "Within tolerance"])
if not interactive:
    print(df)
df.style.format({'Spread': '{:.4%}', 'Hazard rate': '{:.2%}', 'Upfront': '{:.2f}',
                 'Markit value': '{:.2f}', 'Distance': '{:.6f}'})

print('total distance:',distance)



# ---
# jupyter:
#   jupytext:
#     formats: py:light
#     text_representation:
#       extension: .py
#       format_name: light
#       format_version: '1.5'
#       jupytext_version: 1.4.2
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

# # Bermudan swaptions
#
# Copyright (&copy;) 2004, 2005, 2006, 2007 StatPro Italia srl
#
# This file is part of QuantLib, a free-software/open-source library
# for financial quantitative analysts and developers - https://www.quantlib.org/
#
# QuantLib is free software: you can redistribute it and/or modify it under the
# terms of the QuantLib license.  You should have received a copy of the
# license along with this program; if not, please email
# <quantlib-dev@lists.sf.net>. The license is also available online at
# <https://www.quantlib.org/license.shtml>.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the license for more details.

import QuantLib as ql
import pandas as pd

# ### Setup

todaysDate = ql.Date(15, ql.February, 2002)
ql.Settings.instance().evaluationDate = todaysDate
calendar = ql.TARGET()
settlementDate = ql.Date(19, ql.February, 2002)


def calibrate(model, helpers, l, name):
    print("Model: %s" % name)

    method = ql.Simplex(l)
    model.calibrate(helpers, method, ql.EndCriteria(1000, 250, 1e-7, 1e-7, 1e-7))

    print("Parameters: %s" % model.params())

    totalError = 0.0
    data = []
    for swaption, helper in zip(swaptionVols, helpers):
        maturity, length, vol = swaption
        NPV = helper.modelValue()
        implied = helper.impliedVolatility(NPV, 1.0e-4, 1000, 0.05, 0.50)
        error = implied - vol
        totalError += abs(error)
        data.append((maturity, length, vol, implied, error))
    averageError = totalError / len(helpers)

    print(pd.DataFrame(data, columns=["maturity", "length", "volatility", "implied", "error"]))

    print("Average error: %.4f" % averageError)


# ### Market data

swaptionVols = [
    # maturity,          length,             volatility
    (ql.Period(1, ql.Years), ql.Period(5, ql.Years), 0.1148),
    (ql.Period(2, ql.Years), ql.Period(4, ql.Years), 0.1108),
    (ql.Period(3, ql.Years), ql.Period(3, ql.Years), 0.1070),
    (ql.Period(4, ql.Years), ql.Period(2, ql.Years), 0.1021),
    (ql.Period(5, ql.Years), ql.Period(1, ql.Years), 0.1000),
]

# This is a flat yield term structure implying a 1x5 swap at 5%.

rate = ql.makeQuoteHandle(0.04875825)
termStructure = ql.YieldTermStructureHandle(ql.FlatForward(settlementDate, rate, ql.Actual365Fixed()))

# Define the ATM/OTM/ITM swaps:

swapEngine = ql.DiscountingSwapEngine(termStructure)

fixedLegFrequency = ql.Annual
fixedLegTenor = ql.Period(1, ql.Years)
fixedLegConvention = ql.Unadjusted
floatingLegConvention = ql.ModifiedFollowing
fixedLegDayCounter = ql.Thirty360(ql.Thirty360.European)
floatingLegFrequency = ql.Semiannual
floatingLegTenor = ql.Period(6, ql.Months)

payFixed = ql.Swap.Payer
fixingDays = 2
index = ql.Euribor6M(termStructure)
floatingLegDayCounter = index.dayCounter()

swapStart = calendar.advance(settlementDate, 1, ql.Years, floatingLegConvention)
swapEnd = calendar.advance(swapStart, 5, ql.Years, floatingLegConvention)

fixedSchedule = ql.Schedule(
    swapStart,
    swapEnd,
    fixedLegTenor,
    calendar,
    fixedLegConvention,
    fixedLegConvention,
    ql.DateGeneration.Forward,
    False,
)
floatingSchedule = ql.Schedule(
    swapStart,
    swapEnd,
    floatingLegTenor,
    calendar,
    floatingLegConvention,
    floatingLegConvention,
    ql.DateGeneration.Forward,
    False,
)

dummy = ql.VanillaSwap(
    payFixed, 100.0, fixedSchedule, 0.0, fixedLegDayCounter, floatingSchedule, index, 0.0, floatingLegDayCounter
)
dummy.setPricingEngine(swapEngine)
atmRate = dummy.fairRate()

atmSwap = ql.VanillaSwap(
    payFixed, 1000.0, fixedSchedule, atmRate, fixedLegDayCounter,
    floatingSchedule, index, 0.0, floatingLegDayCounter
)

otmSwap = ql.VanillaSwap(
    payFixed, 1000.0, fixedSchedule, atmRate * 1.2, fixedLegDayCounter,
    floatingSchedule, index, 0.0, floatingLegDayCounter
)

itmSwap = ql.VanillaSwap(
    payFixed, 1000.0, fixedSchedule, atmRate * 0.8, fixedLegDayCounter,
    floatingSchedule, index, 0.0, floatingLegDayCounter
)

atmSwap.setPricingEngine(swapEngine)
otmSwap.setPricingEngine(swapEngine)
itmSwap.setPricingEngine(swapEngine)

helpers = [
    ql.SwaptionHelper(
        maturity,
        length,
        ql.makeQuoteHandle(vol),
        index,
        index.tenor(),
        index.dayCounter(),
        index.dayCounter(),
        termStructure,
    )
    for maturity, length, vol in swaptionVols
]

times = {}
for h in helpers:
    for t in h.times():
        times[t] = 1
times = sorted(times.keys())

grid = ql.TimeGrid(times, 30)

G2model = ql.G2(termStructure)
HWmodel = ql.HullWhite(termStructure)
HWmodel2 = ql.HullWhite(termStructure)
BKmodel = ql.BlackKarasinski(termStructure)

# ### Calibrations

for h in helpers:
    h.setPricingEngine(ql.G2SwaptionEngine(G2model, 6.0, 16))
calibrate(G2model, helpers, 0.05, "G2 (analytic formulae)")

for h in helpers:
    h.setPricingEngine(ql.JamshidianSwaptionEngine(HWmodel))
calibrate(HWmodel, helpers, 0.05, "Hull-White (analytic formulae)")

for h in helpers:
    h.setPricingEngine(ql.TreeSwaptionEngine(HWmodel2, grid))
calibrate(HWmodel2, helpers, 0.05, "Hull-White (numerical calibration)")

for h in helpers:
    h.setPricingEngine(ql.TreeSwaptionEngine(BKmodel, grid))
calibrate(BKmodel, helpers, 0.05, "Black-Karasinski (numerical calibration)")


# ### Price Bermudan swaptions on defined swaps

bermudanDates = [d for d in fixedSchedule][:-1]
exercise = ql.BermudanExercise(bermudanDates)

atmSwaption = ql.Swaption(atmSwap, exercise)
otmSwaption = ql.Swaption(otmSwap, exercise)
itmSwaption = ql.Swaption(itmSwap, exercise)

data = []

# +
atmSwaption.setPricingEngine(ql.TreeSwaptionEngine(G2model, 50))
otmSwaption.setPricingEngine(ql.TreeSwaptionEngine(G2model, 50))
itmSwaption.setPricingEngine(ql.TreeSwaptionEngine(G2model, 50))

data.append(("G2 analytic", itmSwaption.NPV(), atmSwaption.NPV(), otmSwaption.NPV()))

# +
atmSwaption.setPricingEngine(ql.TreeSwaptionEngine(HWmodel, 50))
otmSwaption.setPricingEngine(ql.TreeSwaptionEngine(HWmodel, 50))
itmSwaption.setPricingEngine(ql.TreeSwaptionEngine(HWmodel, 50))

data.append(("HW analytic", itmSwaption.NPV(), atmSwaption.NPV(), otmSwaption.NPV()))

# +
atmSwaption.setPricingEngine(ql.TreeSwaptionEngine(HWmodel2, 50))
otmSwaption.setPricingEngine(ql.TreeSwaptionEngine(HWmodel2, 50))
itmSwaption.setPricingEngine(ql.TreeSwaptionEngine(HWmodel2, 50))

data.append(("HW numerical", itmSwaption.NPV(), atmSwaption.NPV(), otmSwaption.NPV()))

# +
atmSwaption.setPricingEngine(ql.TreeSwaptionEngine(BKmodel, 50))
otmSwaption.setPricingEngine(ql.TreeSwaptionEngine(BKmodel, 50))
itmSwaption.setPricingEngine(ql.TreeSwaptionEngine(BKmodel, 50))

data.append(("BK numerical", itmSwaption.NPV(), atmSwaption.NPV(), otmSwaption.NPV()))
# -

print(pd.DataFrame(data, columns=["model", "in-the-money", "at-the-money", "out-of-the-money"]))



"""
 Copyright (C) 2000, 2001, 2002, 2003 RiskMap srl

 This file is part of QuantLib, a free-software/open-source library
 for financial quantitative analysts and developers - http://quantlib.org/

 QuantLib is free software: you can redistribute it and/or modify it
 under the terms of the QuantLib license.  You should have received a
 copy of the license along with this program; if not, please email
 <quantlib-dev@lists.sf.net>. The license is also available online at
 <https://www.quantlib.org/license.shtml>.

 This program is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the license for more details.
"""

import datetime
import operator
import QuantLib as ql
import unittest


class DateTest(unittest.TestCase):
    def testArithmetics(self):
        "Testing date arithmetics"
        today = ql.Date.todaysDate()
        date = today - ql.Period(30, ql.Years)
        end_date = today + ql.Period(30, ql.Years)

        dold = date.dayOfMonth()
        mold = date.month()
        yold = date.year()

        while date < end_date:
            date += 1

            d = date.dayOfMonth()
            m = date.month()
            y = date.year()

            # check if skipping any date
            if not (
                (d == dold + 1 and m == mold and y == yold)
                or (d == 1 and m == mold + 1 and y == yold)
                or (d == 1 and m == 1 and y == yold + 1)
            ):
                self.fail(
                    """
wrong day, month, year increment
    date: %(t)s
    day, month, year: %(d)d, %(m)d, %(y)d
    previous:         %(dold)d, %(mold)d, %(yold)d
                """
                    % locals()
                )
            dold = d
            mold = m
            yold = y

    def test_hash(self):
        for date1 in (ql.Date(1, 2, 2020), ql.Date(3, 4, 2022), ql.Date()):
            for date2 in (ql.Date(1, 2, 2020), ql.Date(3, 4, 2022), ql.Date()):
                expected = str(date1) == str(date2)
                self.assertEqual(date1 == date2, expected)
                self.assertEqual(date1 != date2, not expected)
                self.assertEqual(hash(date1) == hash(date2), expected)

    def test_order(self):
        ops = [operator.eq, operator.ne, operator.lt, operator.le, operator.gt, operator.ge]
        for date1 in (ql.Date(1, 2, 2020), ql.Date(3, 4, 2022), ql.Date()):
            for date2 in (ql.Date(1, 2, 2020), ql.Date(3, 4, 2022), ql.Date()):
                for op in ops:
                    self.assertEqual(op(date1,  date2),
                                     op(date1.serialNumber(), date2.serialNumber()))

    def testHolidayList(self):
        """ Testing Calendar testHolidayList() method. """
        holidayLstFunction = ql.Calendar.holidayList(ql.Poland(), ql.Date(31, 12, 2014), ql.Date(3, 4, 2015), False)
        holidayLstManual = (ql.Date(1, 1, 2015), ql.Date(6, 1, 2015))
        # check if dates both from function and from manual input are the same
        self.assertTrue(all([(a == b) for a, b in zip(holidayLstFunction, holidayLstManual)]))

    def testConversion(self):
        for m in range(1, 13):
            ql_date = ql.Date(m * 2, m, 2020)
            py_date = datetime.date(2020, m, m * 2)
            self.assertEqual(ql_date.to_date(), py_date)
            self.assertEqual(ql.Date.from_date(py_date), ql_date)
            # datetime works as well
            py_dt = datetime.datetime(2020, m, m * 2)
            self.assertEqual(ql.Date.from_date(py_dt), ql_date)

        with self.assertRaisesRegex(RuntimeError, "from_date requires a date"):
            ql.Date.from_date("2020-01-02")


class PeriodTest(unittest.TestCase):
    def test_hash(self):
        for per1 in (ql.Period("1D"), ql.Period("1W"), ql.Period("12M"), ql.Period()):
            for per2 in (ql.Period("1D"), ql.Period("1Y"), ql.Period()):
                expected = str(per1.normalized()) == str(per2.normalized())
                self.assertEqual(per1 == per2, expected)
                self.assertEqual(per1 != per2, not expected)
                self.assertEqual(hash(per1) == hash(per2), expected)

    def test_order(self):
        ops = [operator.eq, operator.ne, operator.lt, operator.le, operator.gt, operator.ge]
        for per1 in (ql.Period("1D"), ql.Period("1W"), ql.Period("12M")):
            for per2 in (ql.Period("1D"), ql.Period("1Y")):
                for op in ops:
                    self.assertEqual(op(per1,  per2), op(per2.frequency(), per1.frequency()))


if __name__ == "__main__":
    print("testing QuantLib", ql.__version__)
    unittest.main(verbosity=2)



# ---
# jupyter:
#   jupytext:
#     formats: py:light
#     text_representation:
#       extension: .py
#       format_name: light
#       format_version: '1.5'
#       jupytext_version: 1.4.2
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

# # Basket options
#
# Copyright (&copy;) 2004, 2005, 2006 StatPro Italia srl
#
# This file is part of QuantLib, a free-software/open-source library
# for financial quantitative analysts and developers - https://www.quantlib.org/
#
# QuantLib is free software: you can redistribute it and/or modify it under the
# terms of the QuantLib license.  You should have received a copy of the
# license along with this program; if not, please email
# <quantlib-dev@lists.sf.net>. The license is also available online at
# <https://www.quantlib.org/license.shtml>.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the license for more details.

import QuantLib as ql

# ### Global data

todaysDate = ql.Date(15, ql.May, 1998)
ql.Settings.instance().evaluationDate = todaysDate
settlementDate = ql.Date(17, ql.May, 1998)
riskFreeRate = ql.FlatForward(settlementDate, 0.05, ql.Actual365Fixed())

# ### Option parameters

exercise = ql.EuropeanExercise(ql.Date(17, ql.May, 1999))
payoff = ql.PlainVanillaPayoff(ql.Option.Call, 8.0)

# ### Market data


underlying1 = ql.SimpleQuote(7.0)
volatility1 = ql.BlackConstantVol(todaysDate, ql.TARGET(), 0.10, ql.Actual365Fixed())
dividendYield1 = ql.FlatForward(settlementDate, 0.05, ql.Actual365Fixed())
underlying2 = ql.SimpleQuote(7.0)
volatility2 = ql.BlackConstantVol(todaysDate, ql.TARGET(), 0.10, ql.Actual365Fixed())
dividendYield2 = ql.FlatForward(settlementDate, 0.05, ql.Actual365Fixed())


process1 = ql.BlackScholesMertonProcess(
    ql.QuoteHandle(underlying1),
    ql.YieldTermStructureHandle(dividendYield1),
    ql.YieldTermStructureHandle(riskFreeRate),
    ql.BlackVolTermStructureHandle(volatility1),
)

process2 = ql.BlackScholesMertonProcess(
    ql.QuoteHandle(underlying2),
    ql.YieldTermStructureHandle(dividendYield2),
    ql.YieldTermStructureHandle(riskFreeRate),
    ql.BlackVolTermStructureHandle(volatility2),
)

matrix = ql.Matrix(2, 2)
matrix[0][0] = 1.0
matrix[1][1] = 1.0
matrix[0][1] = 0.5
matrix[1][0] = 0.5

process = ql.StochasticProcessArray([process1, process2], matrix)

# ### Pricing

basketoption = ql.BasketOption(ql.MaxBasketPayoff(payoff), exercise)
basketoption.setPricingEngine(
    ql.MCEuropeanBasketEngine(process, "pseudorandom", timeStepsPerYear=1, requiredTolerance=0.02, seed=42)
)
print("Maximum Basket Payoff: ", basketoption.NPV())

basketoption = ql.BasketOption(ql.MinBasketPayoff(payoff), exercise)
basketoption.setPricingEngine(
    ql.MCEuropeanBasketEngine(process, "pseudorandom", timeStepsPerYear=1, requiredTolerance=0.02, seed=42)
)
print("Minimum Basket Payoff: ", basketoption.NPV())

basketoption = ql.BasketOption(ql.AverageBasketPayoff(payoff, 2), exercise)
basketoption.setPricingEngine(
    ql.MCEuropeanBasketEngine(process, "pseudorandom", timeStepsPerYear=1, requiredTolerance=0.02, seed=42)
)
print("Average Basket Payoff: ", basketoption.NPV())

americanExercise = ql.AmericanExercise(settlementDate, ql.Date(17, ql.May, 1999))
americanbasketoption = ql.BasketOption(ql.MaxBasketPayoff(payoff), americanExercise)
americanbasketoption.setPricingEngine(
    ql.MCAmericanBasketEngine(
        process,
        "pseudorandom",
        timeSteps=10,
        requiredTolerance=0.02,
        seed=42,
        polynomOrder=5,
        polynomType=ql.LsmBasisSystem.Hermite,
    )
)
print("Basket American Exercise: ", americanbasketoption.NPV())



# ---
# jupyter:
#   jupytext:
#     formats: py:percent
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.4.2
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

# %% [markdown]
# # Global curve bootstrap
#
# Copyright (&copy;) 2020 StatPro Italia srl
#
# This file is part of QuantLib, a free-software/open-source library
# for financial quantitative analysts and developers - https://www.quantlib.org/
#
# QuantLib is free software: you can redistribute it and/or modify it under the
# terms of the QuantLib license.  You should have received a copy of the
# license along with this program; if not, please email
# <quantlib-dev@lists.sf.net>. The license is also available online at
# <https://www.quantlib.org/license.shtml>.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the license for more details.

# %%
import QuantLib as ql
import pandas as pd

# %%
interactive = "get_ipython" in globals()

# %% [markdown]
# ### Setup

# %%
today = ql.Date(26, 9, 2019)
spot = ql.TARGET().advance(today, 2, ql.Days)

# %%
ql.Settings.instance().evaluationDate = today

# %% [markdown]
# ### Data
#
# We'll use the following data as input:

# %%
refMktRates = [
    -0.373,
    -0.388,
    -0.402,
    -0.418,
    -0.431,
    -0.441,
    -0.45,
    -0.457,
    -0.463,
    -0.469,
    -0.461,
    -0.463,
    -0.479,
    -0.4511,
    -0.45418,
    -0.439,
    -0.4124,
    -0.37703,
    -0.3335,
    -0.28168,
    -0.22725,
    -0.1745,
    -0.12425,
    -0.07746,
    0.0385,
    0.1435,
    0.17525,
    0.17275,
    0.1515,
    0.1225,
    0.095,
    0.0644,
]

# %% [markdown]
# ### Market instruments

# %%
index = ql.Euribor6M()

# %% [markdown]
# The first market rate is for the 6-months deposit...

# %%
helpers = [
    ql.DepositRateHelper(
        refMktRates[0] / 100.0, ql.Period(6, ql.Months), 2, ql.TARGET(), ql.ModifiedFollowing, True, ql.Actual360()
    )
]

# %% [markdown]
# ...the next 12 are for FRAs...

# %%
helpers += [ql.FraRateHelper(r / 100.0, i + 1, index) for i, r in enumerate(refMktRates[1:13])]

# %% [markdown]
# ...and the others are swap rates.

# %%
swapTenors = [2, 3, 4, 5, 6, 7, 8, 9, 10, 11, 12, 15, 20, 25, 30, 35, 40, 45, 50]
helpers += [
    ql.SwapRateHelper(
        r / 100.0, ql.Period(T, ql.Years), ql.TARGET(), ql.Annual, ql.ModifiedFollowing, ql.Thirty360(ql.Thirty360.BondBasis), index
    )
    for r, T in zip(refMktRates[13:32], swapTenors)
]

# %% [markdown]
# We'll also add a few synthetic helpers:

# %%
additional_helpers = [ql.FraRateHelper(-0.004, 12 + i, index) for i in range(7)]
additional_dates = [ql.TARGET().advance(spot, 1 + i, ql.Months) for i in range(5)]


# %% [markdown]
# ### Global bootstrap
#
# This curve takes into account the market instruments, as well as the passed additional ones.

# %%
curve = ql.GlobalLinearSimpleZeroCurve(
    spot, helpers, ql.Actual365Fixed(), ql.GlobalBootstrap(additional_helpers, additional_dates, 1.0e-12)
)
curve.enableExtrapolation()


# %% [markdown]
# ### Report

# %%
data = []
for i, h in enumerate(helpers):
    pillar = h.pillarDate()

    if i < 13:
        day_counter = ql.Actual360()
        compounding = ql.Simple
    else:
        day_counter = ql.Thirty360(ql.Thirty360.BondBasis)
        compounding = ql.SimpleThenCompounded

    r = curve.zeroRate(pillar, day_counter, compounding, ql.Annual).rate()
    data.append((pillar.to_date(), r * 100))

# %%
df = pd.DataFrame(data, columns=["pillar", "zero rate"])
if not interactive:
    print(df)
df



# ---
# jupyter:
#   jupytext:
#     formats: py:percent
#     text_representation:
#       extension: .py
#       format_name: percent
#       format_version: '1.3'
#       jupytext_version: 1.4.2
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

# %% [markdown]
# # Leverage function for the Heston SLV model
#
# Copyright (&copy;) 2019 Klaus Spanderen
#
# This file is part of QuantLib, a free-software/open-source library for financial quantitative analysts and developers - https://www.quantlib.org/
#
# QuantLib is free software: you can redistribute it and/or modify it under the
# terms of the QuantLib license.  You should have received a copy of the
# license along with this program; if not, please email
# <quantlib-dev@lists.sf.net>. The license is also available online at
# <https://www.quantlib.org/license.shtml>.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the license for more details.

# %% [markdown]
# This notebook only works with Python 3, at least on Travis.

# %%
import sys

if sys.version_info.major < 3:
    sys.exit()

# %%
import QuantLib as ql
from matplotlib import pyplot as plt
import numpy as np
import math

# %matplotlib inline

# %%
is_interactive = 'get_ipython' in globals()

# %%
todaysDate = ql.Date(15, ql.May, 2019)
ql.Settings.instance().evaluationDate = todaysDate

# %%
settlementDate = todaysDate + ql.Period(2, ql.Days)
exerciseDate = todaysDate + ql.Period(4, ql.Years)

# %%
dc = ql.Actual365Fixed()

spot = 100
underlying = ql.makeQuoteHandle(spot)

riskFreeRate = ql.YieldTermStructureHandle(ql.FlatForward(settlementDate, 0.05, dc))
dividendYield = ql.YieldTermStructureHandle(ql.FlatForward(settlementDate, 0.025, dc))

vol = 0.30
blackVol = ql.BlackVolTermStructureHandle(ql.BlackConstantVol(settlementDate, ql.TARGET(), vol, dc))

# %%
localVol = ql.LocalVolSurface(
    blackVol,
    riskFreeRate,
    dividendYield,
    underlying,
)

hestonProcess = ql.HestonProcess(riskFreeRate, dividendYield, underlying, 0.09, 1.0, 0.06, 0.4, -0.75)

hestonModel = ql.HestonModel(hestonProcess)

# %%
leverageFct = ql.HestonSLVMCModel(
    localVol, hestonModel, ql.MTBrownianGeneratorFactory(1234), exerciseDate, 91
).leverageFunction()

# %%
tSteps = 40
uSteps = 30

tv = np.linspace(0.1, dc.yearFraction(settlementDate, exerciseDate), tSteps)

t = np.empty(tSteps * uSteps)
s = np.empty(tSteps * uSteps)
z = np.empty(tSteps * uSteps)

for i in range(0, tSteps):
    scale = min(4, math.exp(3 * math.sqrt(tv[i]) * vol))
    sv = np.linspace(spot / scale, spot * scale, uSteps)

    for j in range(0, uSteps):
        idx = i * uSteps + j
        t[idx] = tv[i]
        s[idx] = math.log(sv[j])
        z[idx] = leverageFct.localVol(t[idx], sv[j])

# %%
fig = plt.figure(figsize=(12,8))
ax = plt.axes(projection="3d")

surf = ax.plot_trisurf(s, t, z, cmap=plt.cm.viridis, linewidth=0, antialiased=False, edgecolor="none")
ax.view_init(30, -120)

ax.set_xlabel("ln(S)")
ax.set_ylabel("Time")
ax.text2D(0.225, 0.985, "Leverage Function with $\eta=1.0$", transform=ax.transAxes)

fig.colorbar(surf, shrink=0.75, aspect=14)

plt.show(block=False)

# %% [markdown]
# When this is run as a Python script (i.e., from Travis), we need to close the figure in order to terminate.

# %%
if not is_interactive:
    plt.pause(3)
    plt.close()



"""
 Copyright (C) 2000, 2001, 2002, 2003 RiskMap srl

 This file is part of QuantLib, a free-software/open-source library
 for financial quantitative analysts and developers - http://quantlib.org/

 QuantLib is free software: you can redistribute it and/or modify it
 under the terms of the QuantLib license.  You should have received a
 copy of the license along with this program; if not, please email
 <quantlib-dev@lists.sf.net>. The license is also available online at
 <https://www.quantlib.org/license.shtml>.

 This program is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the license for more details.
"""

import QuantLib as ql
import unittest

flag = None


def raiseFlag():
    global flag
    flag = 1


class InstrumentTest(unittest.TestCase):
    def testObservable(self):
        "Testing observability of stocks"
        global flag
        flag = None
        me1 = ql.SimpleQuote(0.0)
        h = ql.RelinkableQuoteHandle(me1)
        s = ql.Stock(h)
        s.NPV()

        obs = ql.Observer(raiseFlag)
        obs.registerWith(s)

        me1.setValue(3.14)
        if not flag:
            self.fail("Observer was not notified of instrument change")

        s.NPV()
        flag = None
        me2 = ql.SimpleQuote(0.0)
        h.linkTo(me2)
        if not flag:
            self.fail("Observer was not notified of instrument change")

        s.NPV()
        flag = None
        s.freeze()
        me2.setValue(2.71)
        if flag:
            self.fail("Observer was notified of frozen instrument change")
        s.unfreeze()
        if not flag:
            self.fail("Observer was not notified of instrument change")


if __name__ == "__main__":
    print("testing QuantLib", ql.__version__)
    unittest.main(verbosity=2)



"""
 Copyright (C) 2000, 2001, 2002, 2003 RiskMap srl
 Copyright (C) 2007 StatPro Italia srl
 Copyright (C) 2020 Marcin Rybacki

 This file is part of QuantLib, a free-software/open-source library
 for financial quantitative analysts and developers - http://quantlib.org/

 QuantLib is free software: you can redistribute it and/or modify it
 under the terms of the QuantLib license.  You should have received a
 copy of the license along with this program; if not, please email
 <quantlib-dev@lists.sf.net>. The license is also available online at
 <https://www.quantlib.org/license.shtml>.

 This program is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the license for more details.
"""

import QuantLib as ql
import unittest
import math

flag = None


def raiseFlag():
    global flag
    flag = 1


def binaryFunction(x, y):
    return 2.0 * x + y


def extrapolatedForwardRate(
        firstSmoothingPoint,
        lastLiquidForwardRate,
        ultimateForwardRate,
        alpha):

    def calculate(t):
        deltaT = t - firstSmoothingPoint
        beta = (1.0 - math.exp(-alpha * deltaT)) / (alpha * deltaT)
        return ultimateForwardRate + (
            lastLiquidForwardRate - ultimateForwardRate) * beta

    return calculate


class TermStructureTest(unittest.TestCase):
    def setUp(self):
        self.calendar = ql.TARGET()
        today = self.calendar.adjust(ql.Date.todaysDate())
        ql.Settings.instance().evaluationDate = today
        self.settlementDays = 2
        self.dayCounter = ql.Actual360()
        self.settlement = self.calendar.advance(today, self.settlementDays, ql.Days)
        deposits = [
            ql.DepositRateHelper(
                ql.makeQuoteHandle(rate / 100),
                ql.Period(n, units),
                self.settlementDays,
                self.calendar,
                ql.ModifiedFollowing,
                False,
                self.dayCounter,
            )
            for (n, units, rate) in [
                (1, ql.Months, 4.581),
                (2, ql.Months, 4.573),
                (3, ql.Months, 4.557),
                (6, ql.Months, 4.496),
                (9, ql.Months, 4.490),
            ]
        ]
        swaps = [
            ql.SwapRateHelper(
                ql.makeQuoteHandle(rate / 100),
                ql.Period(years, ql.Years),
                self.calendar,
                ql.Annual,
                ql.Unadjusted,
                ql.Thirty360(ql.Thirty360.BondBasis),
                ql.Euribor6M(),
            )
            for (years, rate) in [(1, 4.54), (5, 4.99), (10, 5.47), (20, 5.89), (30, 5.96)]
        ]
        self.instruments = deposits + swaps

        self.termStructure = ql.PiecewiseFlatForward(
            self.settlement, self.instruments, self.dayCounter)

    def tearDown(self):
        ql.Settings.instance().evaluationDate = ql.Date()

    def testImpliedObs(self):
        "Testing observability of implied term structure"
        global flag
        flag = None
        h = ql.RelinkableYieldTermStructureHandle()
        settlement = self.termStructure.referenceDate()
        new_settlement = self.calendar.advance(settlement, 3, ql.Years)
        implied = ql.ImpliedTermStructure(h, new_settlement)
        obs = ql.Observer(raiseFlag)
        obs.registerWith(implied)
        h.linkTo(self.termStructure)
        if not flag:
            self.fail("Observer was not notified of term structure change")

    def testFSpreadedObs(self):
        "Testing observability of forward-spreaded term structure"
        global flag
        flag = None
        me = ql.SimpleQuote(0.01)
        mh = ql.QuoteHandle(me)
        h = ql.RelinkableYieldTermStructureHandle()
        spreaded = ql.ForwardSpreadedTermStructure(h, mh)
        obs = ql.Observer(raiseFlag)
        obs.registerWith(spreaded)
        h.linkTo(self.termStructure)
        if not flag:
            self.fail("Observer was not notified of term structure change")
        flag = None
        me.setValue(0.005)
        if not flag:
            self.fail("Observer was not notified of spread change")

    def testZSpreadedObs(self):
        "Testing observability of zero-spreaded term structure"
        global flag
        flag = None
        me = ql.SimpleQuote(0.01)
        mh = ql.QuoteHandle(me)
        h = ql.RelinkableYieldTermStructureHandle()
        spreaded = ql.ZeroSpreadedTermStructure(h, mh)
        obs = ql.Observer(raiseFlag)
        obs.registerWith(spreaded)
        h.linkTo(self.termStructure)
        if not flag:
            self.fail("Observer was not notified of term structure change")
        flag = None
        me.setValue(0.005)
        if not flag:
            self.fail("Observer was not notified of spread change")

    def testCompositeZeroYieldStructure(self):
        """Testing composite zero yield structure"""
        settlement = self.termStructure.referenceDate()
        compounding = ql.Compounded
        freq = ql.Semiannual
        flatTs = ql.FlatForward(
            settlement,
            ql.makeQuoteHandle(0.0085),
            self.dayCounter)
        firstHandle = ql.YieldTermStructureHandle(flatTs)
        secondHandle = ql.YieldTermStructureHandle(self.termStructure)
        compositeTs = ql.CompositeZeroYieldStructure(
            firstHandle, secondHandle, binaryFunction, compounding, freq)
        maturity = settlement + ql.Period(20, ql.Years)
        expectedZeroRate = binaryFunction(
            firstHandle.zeroRate(
                maturity, self.dayCounter, compounding, freq).rate(),
            secondHandle.zeroRate(
                maturity, self.dayCounter, compounding, freq).rate())
        actualZeroRate = compositeTs.zeroRate(
            maturity, self.dayCounter, compounding, freq).rate()
        failMsg = """ Composite zero yield structure rate replication failed:
                        expected zero rate: {expected}
                        actual zero rate: {actual}
                  """.format(expected=expectedZeroRate,
                             actual=actualZeroRate)
        self.assertAlmostEqual(
            first=expectedZeroRate,
            second=actualZeroRate,
            delta=1.0e-12,
            msg=failMsg)

    def testUltimateForwardTermStructure(self):
        """Testing ultimate forward term structure"""
        settlement = self.termStructure.referenceDate()
        ufr = ql.makeQuoteHandle(0.06)
        llfr = ql.makeQuoteHandle(0.05)
        fsp = ql.Period(20, ql.Years)
        alpha = 0.05
        baseCrvHandle = ql.YieldTermStructureHandle(self.termStructure)
        ufrCrv = ql.UltimateForwardTermStructure(
            baseCrvHandle, llfr, ufr, fsp, alpha, None)
        cutOff = ufrCrv.timeFromReference(settlement + fsp)
        forwardCalculator = extrapolatedForwardRate(
            cutOff, llfr.value(), ufr.value(), alpha)
        times = [ufrCrv.timeFromReference(settlement + ql.Period(x, ql.Years))
                 for x in [21, 30, 40, 50, 60, 70, 80, 90, 100]]
        for t in times:
            actualForward = ufrCrv.forwardRate(
                cutOff, t, ql.Continuous, ql.NoFrequency, True).rate()
            expectedForward = forwardCalculator(t)
            failMsg = """ UFR term structure forward replication failed for:
                            time to maturity: {timeToMaturity}
                            expected forward rate: {expected}
                            actual forward rate: {actual}
                      """.format(timeToMaturity=t,
                                 expected=expectedForward,
                                 actual=actualForward)
            self.assertAlmostEqual(
                first=expectedForward,
                second=actualForward,
                delta=1.0e-12,
                msg=failMsg)

    def testTermStructureInterpolationSchemes(self):
        """Testing different interpolation schemes and their consistency"""
        args = [self.settlement, self.instruments, self.dayCounter]
        mapping = [
            [ql.PiecewiseParabolicCubicZero, 
             ql.ParabolicCubicZeroCurve, 'Parabolic Zero'],
            [ql.PiecewiseMonotonicParabolicCubicZero, 
             ql.MonotonicParabolicCubicZeroCurve, 'Monotone Parabolic Zero'],
            [ql.PiecewiseLogParabolicCubicDiscount, 
             ql.LogParabolicCubicDiscountCurve, 'Log Parabolic Discount'],
            [ql.PiecewiseMonotonicLogParabolicCubicDiscount, 
             ql.MonotonicLogParabolicCubicDiscountCurve, 'Monotone Log Parabolic Discount'],
        ]

        for bootstrap, interp, name in mapping:
            bootstrap_crv = bootstrap(*args)
            dates, nodes = zip(*bootstrap_crv.nodes())
            equivalent_crv = interp(dates, nodes, self.dayCounter)

            for d in dates:
                expected = equivalent_crv.zeroRate(
                    d, self.dayCounter, ql.Continuous, ql.NoFrequency).rate()
                actual = bootstrap_crv.zeroRate(
                    d, self.dayCounter, ql.Continuous, ql.NoFrequency).rate()
                
                failMsg = """ Interpolation check failed for:
                            interpolation: {interpolation}
                            expected zero rate: {expected}
                            actual zero rate: {actual}
                      """.format(interpolation=name,
                                 expected=expected,
                                 actual=actual)
                self.assertAlmostEqual(
                    first=expected,
                    second=actual,
                    delta=1.0e-12,
                    msg=failMsg)

    def testTermStructureInterpolationCompatibility(self):
        """Testing different interpolation schemes compatibility"""
        args = (self.settlement, self.instruments, self.dayCounter, ql.IterativeBootstrap())
        types = {
            ql.PiecewiseLogCubicDiscount: [
                ql.LogCubic, ql.MonotonicLogCubic, ql.SplineLogCubic, ql.KrugerLog,
                ql.LogParabolicCubic, ql.MonotonicLogParabolicCubic
            ],
            ql.PiecewiseCubicZero: [
                ql.Cubic, ql.MonotonicCubic, ql.SplineCubic, ql.Kruger,
                ql.ParabolicCubic, ql.MonotonicParabolicCubic
            ],
        }
        for curve_type, inter_types in types.items():
            for interp_type in inter_types:
                curve = curve_type(*args, interp_type())
                self.assertEqual(len(curve.nodes()), len(self.instruments) + 1)

    def testInterpolatedPiecewiseZeroSpreadedTermStructure(self):
        """Testing different interpolation schemes for zero spreaded term structure"""
        h = ql.RelinkableYieldTermStructureHandle()
        h.linkTo(self.termStructure)
        spreads = [(1, 0.005), (2, 0.008), (3, 0.0103), (4, 0.0145), (5, 0.025)]
        dates, quotes = zip(*[(h.referenceDate() + ql.Period(t, ql.Years),
                               ql.QuoteHandle(ql.SimpleQuote(s)))
                              for t, s in spreads])
        args = [h, quotes, dates]
        constructors = [ql.SpreadedLinearZeroInterpolatedTermStructure,
                        ql.SpreadedBackwardFlatZeroInterpolatedTermStructure,
                        ql.SpreadedCubicZeroInterpolatedTermStructure,
                        ql.SpreadedKrugerZeroInterpolatedTermStructure,
                        ql.SpreadedSplineCubicZeroInterpolatedTermStructure]
        for constructor in constructors:
            spreadedTs = constructor(*args)
            for d, r in zip(dates, quotes):
                expected = r.value()

                zeroFromSpreadTS = spreadedTs.zeroRate(
                    d, self.dayCounter, ql.Continuous, ql.NoFrequency).rate()
                zeroFromBaseTs = h.zeroRate(
                    d, self.dayCounter, ql.Continuous, ql.NoFrequency).rate()
                actual = zeroFromSpreadTS - zeroFromBaseTs

                failMsg = """ Interpolated piecewise zero spreaded term structure 
                              zero rate replication failed for:
                            maturity: {maturity}
                            expected zero rate: {expected}
                            actual zero rate: {actual}
                      """.format(maturity=d,
                                 expected=expected,
                                 actual=actual)
                self.assertAlmostEqual(
                    first=expected,
                    second=actual,
                    delta=1.0e-12,
                    msg=failMsg)

    def testQuantoTermStructure(self):
        """Testing quanto term structure"""
        today = ql.Date.todaysDate()

        dividend_ts = ql.YieldTermStructureHandle(
            ql.FlatForward(
                today,
                ql.makeQuoteHandle(0.055),
                self.dayCounter
            )
        )
        r_domestic_ts = ql.YieldTermStructureHandle(
            ql.FlatForward(
                today,
                ql.makeQuoteHandle(-0.01),
                self.dayCounter
            )
        )
        r_foreign_ts = ql.YieldTermStructureHandle(
            ql.FlatForward(
                today,
                ql.makeQuoteHandle(0.02),
                self.dayCounter
            )
        )
        sigma_s = ql.BlackVolTermStructureHandle(
            ql.BlackConstantVol(
                today,
                self.calendar,
                ql.makeQuoteHandle(0.25),
                self.dayCounter
            )
        )
        sigma_fx = ql.BlackVolTermStructureHandle(
            ql.BlackConstantVol(
                today,
                self.calendar,
                ql.makeQuoteHandle(0.05),
                self.dayCounter
            )
        )
        rho = ql.makeQuoteHandle(0.3)
        s_0 = ql.makeQuoteHandle(100.0)

        exercise = ql.EuropeanExercise(self.calendar.advance(today, 6, ql.Months))
        payoff = ql.PlainVanillaPayoff(ql.Option.Call, 95.0)

        vanilla_option = ql.VanillaOption(payoff, exercise)
        quanto_ts = ql.YieldTermStructureHandle(
            ql.QuantoTermStructure(
                dividend_ts,
                r_domestic_ts,
                r_foreign_ts,
                sigma_s,
                ql.nullDouble(),
                sigma_fx,
                ql.nullDouble(),
                rho.value()
            )
        )
        gbm_quanto = ql.BlackScholesMertonProcess(s_0, quanto_ts, r_domestic_ts, sigma_s)
        vanilla_engine = ql.AnalyticEuropeanEngine(gbm_quanto)
        vanilla_option.setPricingEngine(vanilla_engine)

        quanto_option = ql.QuantoVanillaOption(payoff, exercise)
        gbm_vanilla = ql.BlackScholesMertonProcess(s_0, dividend_ts, r_domestic_ts, sigma_s)
        quanto_engine = ql.QuantoEuropeanEngine(gbm_vanilla, r_foreign_ts, sigma_fx, rho)
        quanto_option.setPricingEngine(quanto_engine)

        quanto_option_pv = quanto_option.NPV()
        vanilla_option_pv = vanilla_option.NPV()

        message = """Failed to reproduce QuantoOption / EuropeanQuantoEngine NPV:
                      {quanto_pv}
                      by using the QuantoTermStructure as the dividend together with
                      VanillaOption / AnalyticEuropeanEngine:
                      {vanilla_pv}
                  """.format(
            quanto_pv=quanto_option_pv,
            vanilla_pv=vanilla_option_pv
        )

        self.assertAlmostEqual(
            quanto_option_pv,
            vanilla_option_pv,
            delta=1e-12,
            msg=message
        )

    def testLazyObject(self):
        evaluationDate = ql.Settings.instance().evaluationDate
        nodes = self.termStructure.nodes()
        self.termStructure.freeze()

        ql.Settings.instance().evaluationDate = self.calendar.advance(evaluationDate, 100, ql.Days)

        # Check that dates and rates are unchanged
        for i in range(len(self.termStructure.nodes())):
            self.assertEqual(nodes[i][0], self.termStructure.nodes()[i][0])
            self.assertEqual(nodes[i][1], self.termStructure.nodes()[i][1])

        self.termStructure.recalculate()

        # Check that dates have changed (except the reference, which is fixed)
        for i in range(1, len(self.termStructure.nodes())):
            self.assertNotEqual(nodes[i][0], self.termStructure.nodes()[i][0])

        ql.Settings.instance().evaluationDate = evaluationDate

        self.termStructure.unfreeze()


if __name__ == "__main__":
    print("testing QuantLib", ql.__version__)
    unittest.main(verbosity=2)



import QuantLib as ql
import unittest


class DayCountersTest(unittest.TestCase):
    def test_bus252(self):
        """Test Business252 daycounter"""

        calendar = ql.UnitedStates(ql.UnitedStates.GovernmentBond)

        #
        # Check that SWIG signature for Business252 calendar allows to
        # pass custom calendar into the class constructor.  Old
        # QuantLib-SWIG versions allow only to create Business252
        # calendar with default constructor parameter (Brazil
        # calendar), and generate an exception when trying to pass a
        # custom calendar as a parameter. So we just check here that
        # no exception occurs.
        #
        ql.Business252(calendar)

    def test_hash(self):
        for dc1 in (ql.Actual360(), ql.Thirty365()):
            for dc2 in (ql.Actual360(), ql.Thirty365()):
                expected = dc1.name() == dc2.name()
                self.assertEqual(dc1 == dc2, expected)
                self.assertEqual(dc1 != dc2, not expected)
                self.assertEqual(hash(dc1) == hash(dc2), expected)


if __name__ == "__main__":
    print("testing QuantLib", ql.__version__)
    unittest.main(verbosity=2)



"""
 Copyright (C) 2020 Klaus Spanderen

 This file is part of QuantLib, a free-software/open-source library
 for financial quantitative analysts and developers - http://quantlib.org/

 QuantLib is free software: you can redistribute it and/or modify it
 under the terms of the QuantLib license.  You should have received a
 copy of the license along with this program; if not, please email
 <quantlib-dev@lists.sf.net>. The license is also available online at
 <https://www.quantlib.org/license.shtml>.

 This program is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the license for more details.
"""

import math
import unittest

import QuantLib as ql


class FdmTest(unittest.TestCase):
    def setUp(self):
        self.todaysDate = ql.Date(15, ql.May, 2019)
        ql.Settings.instance().evaluationDate = self.todaysDate

    def tearDown(self):
        ql.Settings.instance().evaluationDate = ql.Date()


    def test1dMesher(self):
        """Testing one dimensional mesher"""

        m = ql.Concentrating1dMesher(0, 1, 10)
        self.assertEqual(m.size(), 10)
        for i in range(0,10):
            self.assertAlmostEqual(m.location(i), i/9.0, 14)

        m = ql.Concentrating1dMesher(0, 1, 10,
            [ql.Concentrating1dMesherPoint(0.75, 0.01,False),
             ql.Concentrating1dMesherPoint(0.5, 0.01, True)])

        self.assertEqual(m.size(), 10)
        self.assertAlmostEqual(m.location(0), 0.0, 14)
        self.assertAlmostEqual(m.location(9), 1.0, 14)

        p = [x for x in m.locations() if ql.close_enough(x, 0.5)]
        self.assertEqual(len(p), 1)
        p = [x for x in m.locations() if ql.close_enough(x, 0.75)]
        self.assertEqual(len(p), 0)

        m = ql.Predefined1dMesher([0,2,4])
        self.assertEqual(m.size(), 3)
        self.assertEqual(m.location(0), 0)
        self.assertEqual(m.location(1), 2)
        self.assertEqual(m.location(2), 4)

    def testFdmLinearOpIterator(self):
        """Testing iterators for linear operators"""

        dim = [2,2,3]
        pos = [0,0,0]
        idx = 0
        opIter = ql.FdmLinearOpIterator(dim, pos, idx)

        self.assertEqual(opIter.index(), 0)

        opIter.increment()
        self.assertEqual(opIter.index(), 1)
        self.assertEqual(opIter.coordinates(), (1, 0, 0))
        opIter.increment()
        self.assertEqual(opIter.coordinates(), (0, 1, 0))

        opIter2 = ql.FdmLinearOpIterator(dim, pos, idx)
        self.assertEqual(opIter.notEqual(opIter2), True)
        self.assertEqual(opIter.notEqual(opIter), False)


    def testFdmLinearOpLayout(self):
        """Testing memory layout for linear operators"""

        dim = [2,2,3]

        m = ql.FdmLinearOpLayout(dim)

        self.assertEqual(m.size(), 2*2*3)
        self.assertEqual(m.dim(), (2, 2, 3))
        self.assertEqual(m.spacing(), (1, 2, 4))
        self.assertEqual(m.index((0,1,2)), 10)
        self.assertEqual(m.neighbourhood(m.begin(), 0, 1), 1)
        self.assertEqual(m.neighbourhood(m.begin(), 2, 2), 8)
        self.assertEqual(m.neighbourhood(m.begin(), 0, 1, 2, 2), 9)

        n = m.iter_neighbourhood(m.begin(), 0, 1)
        opIter = m.begin()
        opIter.increment()

        self.assertEqual(opIter.notEqual(n), False)

    def testFdmMesherComposite(self):
        """Testing mesher composites"""

        m1 = ql.Concentrating1dMesher(0, 1, 2)
        m2 = ql.Uniform1dMesher(0, 2, 3)

        m = ql.FdmMesherComposite(m1, m2)
        self.assertEqual(len(m.getFdm1dMeshers()), 2)

        locations = m.locations(0)
        self.assertEqual(len(locations), 6)

        self.assertEqual(list(map(round, locations)), [0, 1, 0, 1, 0, 1])

        locations = m.locations(1)
        self.assertEqual(list(map(round, locations)), [0, 0, 1, 1, 2, 2])

    def testFdmLinearOpComposite(self):
        """Testing linear operator composites"""

        class Foo:
            t1 = 0.0
            t2 = 0.0

            @classmethod
            def size(self):
                return 42

            def setTime(self, t1, t2):
                self.t1 = t1
                self.t2 = t2

            @classmethod
            def apply(self, r):
                return 2*r

            @classmethod
            def apply_mixed(self, r):
                return 3*r

            @classmethod
            def apply_direction(self, direction , r):
                return direction*r

            @classmethod
            def solve_splitting(self, direction , r, s):
                return direction*s*r

            @classmethod
            def preconditioner(self, r, s):
                return s*r


        foo = Foo()

        c = ql.FdmLinearOpCompositeProxy(foo)

        self.assertEqual(c.size(), foo.size())

        c.setTime(1.0, 2.0)
        self.assertAlmostEqual(foo.t1, 1.0, 14)
        self.assertAlmostEqual(foo.t2, 2.0, 14)

        r = ql.Array([1,2,3,4])
        self.assertEqual(c.apply(r), 2*r)
        self.assertEqual(c.apply_mixed(r), 3*r)
        self.assertEqual(c.apply_direction(7, r), 7*r)

        s = c.solve_splitting(7, r, 0.5)
        self.assertEqual(len(s), len(r))
        for i, x in enumerate(s):
            self.assertAlmostEqual(x, 3.5*r[i], 14)

        self.assertEqual(c.preconditioner(r, 4), 4*r)

        class Bar:
            @classmethod
            def apply(self, r):
                return 1

            def apply_mixed(self, r):
                pass

        with self.assertRaises(RuntimeError):
            ql.FdmLinearOpCompositeProxy(Bar()).apply(r)

        with self.assertRaises(RuntimeError):
            ql.FdmLinearOpCompositeProxy(Bar()).apply_mixed(r)


    def testFdmBlackScholesOp(self):
        """Testing linear Black-Scholes operator"""

        todaysDate = ql.Date(1, ql.January, 2020)
        ql.Settings.instance().evaluationDate = todaysDate
        dc = ql.Actual365Fixed()

        settlementDate = todaysDate + 2
        riskFreeRate = ql.FlatForward(settlementDate, 0.05, dc)

        exercise = ql.EuropeanExercise(ql.Date(27, ql.December, 2020))
        maturity = dc.yearFraction(todaysDate, exercise.lastDate())

        strike = 110.0
        payoff = ql.PlainVanillaPayoff(ql.Option.Call, strike)

        underlying = ql.SimpleQuote(100.0)
        volatility = ql.BlackConstantVol(settlementDate, ql.TARGET(), 0.10, dc)
        dividendYield = ql.FlatForward(settlementDate, 0.05, dc)

        process = ql.BlackScholesMertonProcess(
            ql.QuoteHandle(underlying),
            ql.YieldTermStructureHandle(dividendYield),
            ql.YieldTermStructureHandle(riskFreeRate),
            ql.BlackVolTermStructureHandle(volatility)
        )

        mesher = ql.FdmMesherComposite(
            ql.FdmBlackScholesMesher(10, process, maturity, strike))

        op = ql.FdmBlackScholesOp(mesher, process, strike)
        self.assertEqual(op.size(), 1)

        op.setTime(0, 0.1)

        c = [payoff(math.exp(x)) for x in mesher.locations(0)]
        p = op.apply(c)

        e = [ 0.0, 0.0, 0.0, 0.0, 0.0,
              3.18353, 0.755402, -1.30583, -2.19881, -4.0271 ]

        for i, x in enumerate(e):
            self.assertAlmostEqual(x, p[i], 5)


    def testFdmFirstOrderOperator(self):
        """Testing first order operator"""

        mesher = ql.Uniform1dMesher(0.0, math.pi, 1000)

        op = ql.FirstDerivativeOp(0, ql.FdmMesherComposite(mesher))

        l = mesher.locations()

        x = list(map(math.sin, l))

        y = op.apply(x)

        for u, v in zip(l, y):
            self.assertAlmostEqual(v, math.cos(u), 4)


    def testFdmSecondOrderOperator(self):
        """Testing second order operator"""

        mesher = ql.Uniform1dMesher(0.0, math.pi, 1000)

        op = ql.SecondDerivativeOp(0, ql.FdmMesherComposite(mesher))

        x = list(map(math.sin, mesher.locations()))

        y = op.apply(x)

        for u, v in zip(x, y):
            self.assertAlmostEqual(v, -u, 4)

    def testFdmBoundaryCondition(self):
        """Testing Dirichlet Boundary conditions"""

        m = ql.FdmMesherComposite(
            ql.Uniform1dMesher(0.0, 1.0, 5))

        b = ql.FdmDirichletBoundary(
            m, math.pi, 0, ql.FdmBoundaryCondition.Upper)

        x = ql.Array(len(m.locations(0)), 0.0)

        b.applyAfterApplying(x)

        self.assertEqual(list(x), [0,0,0,0, math.pi])

        s = ql.FdmBoundaryConditionSet()
        s.push_back(b)

        self.assertEqual(len(s), 1)

    def testFdmStepConditionCallBack(self):
        """Testing step condition call back function"""

        class Foo:
            @classmethod
            def applyTo(self, a, t):
                for i in range(5):
                    a[i] = t+1.0

        m = ql.FdmStepConditionProxy(Foo())

        x = ql.Array(5)

        m.applyTo(x, 2.0)

        self.assertEqual(len(x), 5)
        self.assertEqual(list(x), [3.0, 3.0, 3.0, 3.0, 3.0])

    def testFdmInnerValueCalculatorCallBack(self):
        """Testing inner value call back function"""

        class Foo:
            @classmethod
            def innerValue(self, opIter, t):
                return opIter.index() + t

            @classmethod
            def avgInnerValue(self, opIter, t):
                return opIter.index() + 2*t

        m = ql.FdmInnerValueCalculatorProxy(Foo())

        dim = [2,2,3]
        pos = [0,0,0]

        opIter = ql.FdmLinearOpIterator(dim, pos, 0)

        while (opIter.index() < 2*2*3):
            idx = opIter.index()

            self.assertEqual(m.innerValue(opIter, 2.0), idx + 2.0)
            self.assertEqual(m.avgInnerValue(opIter, 2.0), idx + 4.0)

            opIter.increment()


    def testFdmLogInnerValueCalculator(self):
        """Testing log inner value calculator"""

        m = ql.FdmMesherComposite(
            ql.Uniform1dMesher(math.log(50), math.log(150), 11))

        p = ql.PlainVanillaPayoff(ql.Option.Call, 100)

        v = ql.FdmLogInnerValue(p, m, 0)

        opIter = m.layout().begin()
        while opIter.notEqual(m.layout().end()):
            x = math.exp(m.location(opIter, 0));
            self.assertAlmostEqual(p(x), v.innerValue(opIter, 1.0), 14)
            opIter.increment()


    def testAmericanOptionPricing(self):
        """Testing Black-Scholes and Heston American Option pricing"""

        xSteps = 100
        tSteps = 25
        dampingSteps = 0

        todaysDate = ql.Date(15, ql.January, 2020)
        ql.Settings.instance().evaluationDate = todaysDate

        dc = ql.Actual365Fixed()

        riskFreeRate = ql.YieldTermStructureHandle(
            ql.FlatForward(todaysDate, 0.06, dc))
        dividendYield = ql.YieldTermStructureHandle(
            ql.FlatForward(todaysDate, 0.02, dc))

        strike = 110.0
        payoff = ql.PlainVanillaPayoff(ql.Option.Put, strike)

        maturityDate = todaysDate + ql.Period(1, ql.Years)
        maturity = dc.yearFraction(todaysDate, maturityDate)

        exercise = ql.AmericanExercise(todaysDate, maturityDate)

        spot = ql.makeQuoteHandle(100.0)
        volatility = ql.BlackConstantVol(todaysDate, ql.TARGET(), 0.20, dc)

        process = ql.BlackScholesMertonProcess(
            spot, dividendYield, riskFreeRate,
            ql.BlackVolTermStructureHandle(volatility)
        )

        option = ql.VanillaOption(payoff, exercise)
        option.setPricingEngine(ql.FdBlackScholesVanillaEngine.make(
            process, xGrid = xSteps, tGrid = tSteps,
            dampingSteps = dampingSteps)
        )

        expected = option.NPV()

        equityMesher = ql.FdmBlackScholesMesher(
            xSteps, process, maturity,
            strike, cPoint = (strike, 0.1)
        )

        mesher = ql.FdmMesherComposite(equityMesher)

        op = ql.FdmBlackScholesOp(mesher, process, strike)

        innerValueCalculator = ql.FdmLogInnerValue(payoff, mesher, 0)

        x = []
        rhs = []
        layout = mesher.layout()
        opIter = layout.begin()
        while (opIter.notEqual(layout.end())):
            x.append(mesher.location(opIter, 0))
            rhs.append(innerValueCalculator.avgInnerValue(opIter, maturity))
            opIter.increment()

        rhs = ql.Array(rhs)

        bcSet = ql.FdmBoundaryConditionSet()
        stepCondition = ql.FdmStepConditionComposite.vanillaComposite(
            ql.DividendSchedule(), exercise, mesher,
            innerValueCalculator, todaysDate, dc
        )

        # only to test an Operator defined in python
        class OperatorProxy:
            def __init__(self, op):
                self.op = op

            def size(self):
                return self.op.size()

            def setTime(self, t1, t2):
                return self.op.setTime(t1, t2)

            def apply(self, r):
                return self.op.apply(r)

            def apply_direction(self, i, r):
                return self.op.apply_direction(i, r)

            def solve_splitting(self, i, r, s):
                return self.op.solve_splitting(i, r, s)


        proxyOp = ql.FdmLinearOpCompositeProxy(OperatorProxy(op))

        solver = ql.FdmBackwardSolver(
            proxyOp, bcSet, stepCondition, ql.FdmSchemeDesc.Douglas()
        )

        solver.rollback(rhs, maturity, 0.0, tSteps, dampingSteps)

        spline = ql.CubicNaturalSpline(x, rhs);

        logS = math.log(spot.value())

        calculated = spline(logS)

        self.assertAlmostEqual(calculated, expected, 1)

        solverDesc = ql.FdmSolverDesc(
            mesher, bcSet, stepCondition, innerValueCalculator,
            maturity, tSteps, dampingSteps)

        calculated = ql.Fdm1DimSolver(
            solverDesc, ql.FdmSchemeDesc.Douglas(), op).interpolateAt(logS)

        self.assertAlmostEqual(calculated, expected, 2)

        v0 = 0.4*0.4
        kappa = 1.0
        theta = v0
        sigma = 1e-4
        rho = 0.0

        hestonProcess = ql.HestonProcess(
            riskFreeRate, dividendYield,
            spot, v0, kappa, theta, sigma, rho)

        leverageFct = ql.LocalVolSurface(
            ql.BlackVolTermStructureHandle(
                ql.BlackConstantVol(todaysDate, ql.TARGET(), 0.50, dc)),
            riskFreeRate,
            dividendYield,
            spot.value()
        )

        vSteps = 3

        vMesher = ql.FdmHestonLocalVolatilityVarianceMesher(
            vSteps, hestonProcess, leverageFct, maturity)

        avgVolaEstimate = vMesher.volaEstimate()

        self.assertAlmostEqual(avgVolaEstimate, 0.2, 5)

        mesher = ql.FdmMesherComposite(equityMesher, vMesher)

        innerValueCalculator = ql.FdmLogInnerValue(payoff, mesher, 0)

        stepCondition = ql.FdmStepConditionComposite.vanillaComposite(
            ql.DividendSchedule(), exercise, mesher,
            innerValueCalculator, todaysDate, dc
        )

        solverDesc = ql.FdmSolverDesc(
            mesher, bcSet, stepCondition, innerValueCalculator,
            maturity, tSteps, dampingSteps)

        calculated = ql.FdmHestonSolver(
            hestonProcess, solverDesc, leverageFct = leverageFct).valueAt(
                spot.value(), 0.16)

        self.assertAlmostEqual(calculated, expected, 1)


    def testBSMRNDCalculator(self):
        """Testing Black-Scholes risk neutral density calculator"""

        dc = ql.Actual365Fixed()
        todaysDate = ql.Date(15, ql.January, 2020)

        r   = 0.0
        q   = 0.0
        vol = 0.2
        s0  = 100

        process = ql.BlackScholesMertonProcess(
            ql.makeQuoteHandle(s0),
            ql.YieldTermStructureHandle(
                ql.FlatForward(todaysDate, q, dc)),
            ql.YieldTermStructureHandle(
                ql.FlatForward(todaysDate, r, dc)),
            ql.BlackVolTermStructureHandle(
                ql.BlackConstantVol(todaysDate, ql.TARGET(), vol, dc))
        )

        rnd = ql.BSMRNDCalculator(process)

        t = 1.2
        x = math.log(80.0)

        mu = math.log(s0) + (r-q-0.5*vol*vol)*t

        calculated = rnd.pdf(x, t)

        stdev = vol * math.sqrt(t)

        expected = (1.0/(math.sqrt(2*math.pi)*stdev) *
            math.exp( -0.5*math.pow((x-mu)/stdev, 2.0) ))

        self.assertAlmostEqual(calculated, expected, 8)


    def testOrnsteinUhlenbeckVsBachelier(self):
        """Testing Fdm Ornstein-Uhlenbeck pricing"""

        todaysDate = ql.Date(15, ql.January, 2020)
        ql.Settings.instance().evaluationDate = todaysDate

        dc = ql.Actual365Fixed()

        rTS = ql.FlatForward(todaysDate, 0.06, dc)

        strike = 110.0
        payoff = ql.PlainVanillaPayoff(ql.Option.Put, strike)

        maturityDate = todaysDate + ql.Period(2, ql.Years)

        exercise = ql.EuropeanExercise(maturityDate)

        option = ql.VanillaOption(payoff, exercise)

        x0 = 100
        sigma = 20.0
        speed = 5

        pdeEngine = ql.FdOrnsteinUhlenbeckVanillaEngine(
            ql.OrnsteinUhlenbeckProcess(speed, sigma, x0, x0), rTS, 50
        )

        option.setPricingEngine(pdeEngine)
        calculated = option.NPV()

        stdev = math.sqrt(sigma*sigma/(2*speed))

        expected = ql.bachelierBlackFormula(
            ql.Option.Put,
            strike, x0, stdev,
            rTS.discount(maturityDate)
        )

        self.assertAlmostEqual(calculated, expected, 2)


    def testSparseLinearMatrixSolver(self):
        """Testing sparse linear matrix solver"""

        A = ql.Matrix([
            [1.0, 0.0, 1.0],
            [0.0, 1.0, 0.5],
            [1.0, 0.5, 1.0]
        ])

        b = ql.Array([ 1.0, 0.2, 0.5 ])

        expected = ql.inverse(A)*b

        def foo(x):
            return A*x

        calculated = ql.BiCGstab(
            ql.MatrixMultiplicationProxy(foo), 100, 1e-6).solve(b)

        for i in range(3):
            self.assertAlmostEqual(expected[i], calculated[i], 4)

        calculated = ql.GMRES(
            ql.MatrixMultiplicationProxy(foo), 100, 1e-6).solve(b)

        for i in range(3):
            self.assertAlmostEqual(expected[i], calculated[i], 4)

        def preconditioner(x):
            return ql.inverse(A)*x

        calculated = ql.BiCGstab(
            ql.MatrixMultiplicationProxy(foo), 100, 1e-6,
            ql.MatrixMultiplicationProxy(preconditioner)).solve(b)

        for i in range(3):
            self.assertAlmostEqual(expected[i], calculated[i], 4)

    def testGlued1dMesher(self):
        """Testing sparse linear matrix solver"""

        m1 = ql.Uniform1dMesher(0, 2, 3)
        m2 = ql.Uniform1dMesher(2, 4, 3)

        m3 = ql.Glued1dMesher(m1, m2)

        self.assertEqual(m3.locations(), (0,1,2,3,4))

    def testFdmZeroInnerValue(self):
        """Testing FdmZeroInnerValue"""
        opIter = ql.FdmLinearOpIterator([1], [0], 0)

        self.assertEqual(ql.FdmZeroInnerValue().innerValue(opIter, 1.0), 0.0)


if __name__ == "__main__":
    print("testing QuantLib", ql.__version__)
    unittest.main(verbosity=2)



import sysconfig
import concurrent.futures
import time

import QuantLib as ql

if __name__ == "__main__":
    def pricing() -> float:
        todaysDate = ql.Date(15, ql.May, 2025)
        ql.Settings.instance().evaluationDate = todaysDate
        underlying = ql.SimpleQuote(7.0)
        dividendYield = ql.FlatForward(todaysDate, 0.05, ql.Actual365Fixed())
        volatility = ql.BlackConstantVol(todaysDate, ql.TARGET(), 0.10, ql.Actual365Fixed())
        riskFreeRate = ql.FlatForward(todaysDate, 0.05, ql.Actual365Fixed())

        process = ql.BlackScholesMertonProcess(
            ql.QuoteHandle(underlying),
            ql.YieldTermStructureHandle(dividendYield),
            ql.YieldTermStructureHandle(riskFreeRate),
            ql.BlackVolTermStructureHandle(volatility),
        )
        option = ql.VanillaOption(
            ql.PlainVanillaPayoff(ql.Option.Call, 8.0),
            ql.EuropeanExercise(ql.Date(17, ql.May, 2026))
        )
        engine = ql.FdBlackScholesVanillaEngine(process, 100, 8000)
        option.setPricingEngine(engine)
        return option.NPV()

    print("GIL is disabled: ", bool(sysconfig.get_config_var("Py_GIL_DISABLED")))

    for w in [1, 2, 4, 8, 16, 32, 64]:
        start = time.time()

        with concurrent.futures.ThreadPoolExecutor(max_workers=w) as executor:
            retVals = [executor.submit(pricing) for i in range(400)]

            for val in retVals:
                val.result()

        print(f"{w:2} threads, {400/(time.time()-start):5.1f} NPVs per sec")



"""
 Copyright (C) 2025 Hiroto Ogawa

 This file is part of QuantLib, a free-software/open-source library
 for financial quantitative analysts and developers - http://quantlib.org/

 QuantLib is free software: you can redistribute it and/or modify it
 under the terms of the QuantLib license.  You should have received a
 copy of the license along with this program; if not, please email
 <quantlib-dev@lists.sf.net>. The license is also available online at
 <https://www.quantlib.org/license.shtml>.

 This program is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the license for more details.
"""
import unittest
import math
import QuantLib as ql

LAG = 0
DC = ql.ActualActual(ql.ActualActual.ISDA)
CAL = ql.NullCalendar()

def flat_rate(rate):
    return ql.FlatForward(
        LAG, CAL, ql.makeQuoteHandle(rate), DC)

class PerpetualFuturesTest(unittest.TestCase):
    def test_perpetual_futures(self):
        """ Testing PerpetualFutures pricing by discounted cashflow method against analytic formulae. """
        val_date = ql.Date(20, 6, 2025)
        ql.Settings.instance().evaluationDate = val_date

        test_dicts = [
            # Discrete time
            {"payoff": ql.PerpetualFutures.Linear, "funding": ql.PerpetualFutures.FundingWithPreviousSpot,
             "freq": ql.Period(3, ql.Months), "spot": 10000., "domRate": 0.04, "btcYield": 0.02,
             "fundingRate": 0.01, "rateDiff": 0.005, "tol": 1.e-6},
            {"payoff": ql.PerpetualFutures.Linear, "funding": ql.PerpetualFutures.FundingWithCurrentSpot,
             "freq": ql.Period(3, ql.Months), "spot": 10000., "domRate": 0.04, "btcYield": 0.02,
             "fundingRate": 0.01, "rateDiff": 0.005, "tol": 1.e-6},
            {"payoff": ql.PerpetualFutures.Inverse, "funding": ql.PerpetualFutures.FundingWithPreviousSpot,
             "freq": ql.Period(3, ql.Months), "spot": 10000., "domRate": 0.04, "btcYield": 0.02,
             "fundingRate": 0.01, "rateDiff": 0.005, "tol": 1.e-6},
            {"payoff": ql.PerpetualFutures.Inverse, "funding": ql.PerpetualFutures.FundingWithCurrentSpot,
             "freq": ql.Period(3, ql.Months), "spot": 10000., "domRate": 0.04, "btcYield": 0.02,
             "fundingRate": 0.01, "rateDiff": 0.005, "tol": 1.e-6},
             # Continuous time
            {"payoff": ql.PerpetualFutures.Linear, "funding": ql.PerpetualFutures.FundingWithPreviousSpot,
             "freq": ql.Period(0, ql.Months), "spot": 10000., "domRate": 0.04, "btcYield": 0.02,
             "fundingRate": 0.2, "rateDiff": 0.005, "tol": 1.e-6},
            {"payoff": ql.PerpetualFutures.Inverse, "funding": ql.PerpetualFutures.FundingWithPreviousSpot,
             "freq": ql.Period(0, ql.Months), "spot": 10000., "domRate": 0.04, "btcYield": 0.02,
             "fundingRate": 0.2, "rateDiff": 0.005, "tol": 1.e-6},
        ]

        for d in test_dicts:
            pf = ql.PerpetualFutures(d["payoff"], d["funding"], d["freq"], CAL, DC)
            domYC = ql.YieldTermStructureHandle(flat_rate(d["domRate"]))
            btcYC = ql.YieldTermStructureHandle(flat_rate(d["btcYield"]))
            spot = ql.QuoteHandle(ql.SimpleQuote(d["spot"]))

            fundingTimes = [0.0]
            fundingRates = [d["fundingRate"]]
            ir_diffs = [d["rateDiff"]]
            engine = ql.DiscountingPerpetualFuturesEngine(domYC, btcYC, spot, fundingTimes, fundingRates, ir_diffs,
                                                      ql.DiscountingPerpetualFuturesEngine.PiecewiseConstant)
            # pricing
            pf.setPricingEngine(engine)
            npv = pf.NPV()

            # analytic price
            # for details, refer to
            # Perpetual Futures Pricing, Damien Ackerer, Julien Hugonnier, Urban Jermann, 2024
            # https://finance.wharton.upenn.edu/~jermann/AHJ-main-10.pdf
            period = d["freq"]
            length = float(period.length())
            unit = period.units()
            if unit == ql.Years:
                dt = length
            elif unit == ql.Months:
                dt = length / 12.
            elif unit == ql.Weeks:
                dt = length * 7. / 365.
            elif unit == ql.Days:
                dt = length / 365.
            elif unit == ql.Hours:
                dt = length / 365. / 24.
            elif unit == ql.Minutes:
                dt = length / 365. / 24. / 60.
            elif unit == ql.Seconds:
                dt = length / 365. / 24. / 60. / 60.
            else:
                raise RuntimeError("unknown funding frequency unit: " + str(unit))
            
            # Discrete time
            if length > 0:
                if d["payoff"] == ql.PerpetualFutures.Linear:
                    if d["funding"] == ql.PerpetualFutures.FundingWithPreviousSpot:
                        # Equation (12) in the above paper
                        expected = (
                            d["spot"] * (d["fundingRate"] - d["rateDiff"]) * math.exp(d["btcYield"] * dt) /
                            (math.exp(d["btcYield"] * dt) - math.exp(d["domRate"] * dt) + d["fundingRate"] * math.exp(d["btcYield"] * dt)))
                    elif d["funding"] == ql.PerpetualFutures.FundingWithCurrentSpot:
                        # at the end of "3 Perpetual futures pricing" in the above paper
                        expected = (
                            d["spot"] * (d["fundingRate"] - d["rateDiff"]) * math.exp(d["domRate"] * dt) /
                            (math.exp(d["btcYield"] * dt) - math.exp(d["domRate"] * dt) + d["fundingRate"] * math.exp(d["domRate"] * dt)))
                elif d["payoff"] == ql.PerpetualFutures.Inverse:
                    if d["funding"] == ql.PerpetualFutures.FundingWithPreviousSpot:
                        # "Proposition 2" in the above paper
                        expected = (
                            d["spot"] *
                            (math.exp(d["domRate"] * dt) - math.exp(d["btcYield"] * dt) + d["fundingRate"] * math.exp(d["domRate"] * dt)) /
                            (d["fundingRate"] - d["rateDiff"]) / math.exp(d["domRate"] * dt))
                    elif d["funding"] == ql.PerpetualFutures.FundingWithCurrentSpot:
                        expected = (
                        d["spot"] *
                        (math.exp(d["domRate"] * dt) - math.exp(d["btcYield"] * dt) + d["fundingRate"] * math.exp(d["btcYield"] * dt)) /
                        (d["fundingRate"] - d["rateDiff"]) / math.exp(d["btcYield"] * dt))
            else:
                # Continuous time
                if d["payoff"] == ql.PerpetualFutures.Linear:
                    # "Proposition 3" in the above paper
                    expected = d["spot"] * (d["fundingRate"] - d["rateDiff"]) / (d["btcYield"] - d["domRate"] + d["fundingRate"])
                elif d["payoff"] == ql.PerpetualFutures.Inverse:
                    # "Proposition 4" in the above paper
                    expected = d["spot"] * (d["domRate"] - d["btcYield"] + d["fundingRate"]) / (d["fundingRate"] - d["rateDiff"])
            failed_msg = f"Perpetual future price {npv} differs from analytic price {expected} with relative tolerance {d['tol']}\n"
            self.assertAlmostEqual(npv, expected, delta=d["tol"] * expected, msg=failed_msg)

if __name__ == '__main__':
    print("testing QuantLib", ql.__version__)
    unittest.main(verbosity=2)



"""
 Copyright (C) 2020 Marcin Rybacki

 This file is part of QuantLib, a free-software/open-source library
 for financial quantitative analysts and developers - http://quantlib.org/

 QuantLib is free software: you can redistribute it and/or modify it
 under the terms of the QuantLib license.  You should have received a
 copy of the license along with this program; if not, please email
 <quantlib-dev@lists.sf.net>. The license is also available online at
 <https://www.quantlib.org/license.shtml>.

 This program is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the license for more details.
"""

import unittest
import QuantLib as ql


BASIS_POINT = 1e-4

EPSILON = 1.e-10

OPTION_TYPE_MAP = {ql.Swap.Receiver: 'Receiver',
                   ql.Swap.Payer: 'Payer'}

SETTLEMENT_TYPE_MAP = {ql.Settlement.Physical: 'Physical',
                       ql.Settlement.Cash: 'Cash'}

SETTLEMENT_METHOD_MAP = {ql.Settlement.PhysicalOTC: 'Physical OTC',
                         ql.Settlement.CollateralizedCashPrice: (
                             'Collateralized Cash Price'),
                         ql.Settlement.ParYieldCurve: 'Par Yield Curve'}


def compounded_annual_constant_rate_discount(
        rate,
        day_counter):
    def _calc(start, end):
        time = day_counter.yearFraction(start, end)
        return (1.0 + rate) ** (-time)
    return _calc


def par_yield_bps(underlying,
                  discount_handle):
    fixed_leg = underlying.fixedLeg()
    first_coupon = ql.as_fixed_rate_coupon(fixed_leg[0])
    discount_date = first_coupon.accrualStartDate()
    discount = discount_handle.discount(discount_date)
    fixed_rate = underlying.fixedRate()
    fixed_dct = underlying.fixedDayCount()
    fair_rate = underlying.fairRate()
    ir_func = compounded_annual_constant_rate_discount(fair_rate, fixed_dct)
    bps = sum([ir_func(discount_date, c_f.date()) * c_f.amount() / fixed_rate
               for c_f in fixed_leg
               if c_f.date() > discount_date])
    return abs(bps) * discount


def swap_pv01(underlying):
    return abs(underlying.fixedLegBPS()) / BASIS_POINT


def make_const_black_vol_engine(discount_handle, volatility):
    h = ql.makeQuoteHandle(volatility)
    return ql.BlackSwaptionEngine(discount_handle, h)


def make_const_bachelier_vol_engine(discount_handle, volatility):
    h = ql.makeQuoteHandle(volatility)
    return ql.BachelierSwaptionEngine(discount_handle, h)


class SwaptionTest(unittest.TestCase):
    def setUp(self):
        self.calendar = ql.TARGET()
        self.today = self.calendar.adjust(ql.Date.todaysDate())
        ql.Settings.instance().evaluationDate = self.today

        projection_curve_handle = ql.RelinkableYieldTermStructureHandle()
        self.projection_rate = 0.01
        self.projection_quote_handle = ql.RelinkableQuoteHandle()
        projection_curve = ql.FlatForward(
            self.today, self.projection_quote_handle, ql.Actual365Fixed())
        projection_curve_handle.linkTo(projection_curve)

        self.discount_handle = ql.YieldTermStructureHandle(ql.FlatForward(
            self.today, ql.makeQuoteHandle(0.0085), ql.Actual365Fixed()))
        self.swap_engine = ql.DiscountingSwapEngine(self.discount_handle)

        self.idx = ql.Euribor6M(projection_curve_handle)

        self.exercises = [ql.Period(1, ql.Years), ql.Period(2, ql.Years),
                          ql.Period(3, ql.Years), ql.Period(5, ql.Years),
                          ql.Period(7, ql.Years), ql.Period(10, ql.Years)]
        self.lengths = [ql.Period(1, ql.Years), ql.Period(2, ql.Years),
                        ql.Period(3, ql.Years), ql.Period(5, ql.Years),
                        ql.Period(7, ql.Years), ql.Period(10, ql.Years),
                        ql.Period(15, ql.Years), ql.Period(20, ql.Years)]
        self.swap_type = [ql.Swap.Receiver, ql.Swap.Payer]

    def tearDown(self):
        ql.Settings.instance().evaluationDate = ql.Date()

    def _assert_swaption_annuity(self,
                                 swaption_pricer_func,
                                 use_bachelier_vol):
        self.projection_quote_handle.linkTo(
            ql.SimpleQuote(self.projection_rate))

        settle_type = ql.Settlement.Cash
        methods = [ql.Settlement.ParYieldCurve,
                   ql.Settlement.CollateralizedCashPrice]

        for e in self.exercises:
            for l in self.lengths:
                for t in self.swap_type:
                    for m in methods:
                        volatility = 0.003 if use_bachelier_vol else 0.3
                        strike = 0.03
                        swaption_engine = swaption_pricer_func(
                            self.discount_handle, volatility)
                        exercise_date = self.calendar.advance(
                            self.today, e)
                        start_date = self.calendar.advance(
                            exercise_date, ql.Period(2, ql.Days))

                        underlying = ql.MakeVanillaSwap(
                            l, self.idx, strike, ql.Period(0, ql.Days),
                            effectiveDate=start_date,
                            fixedLegTenor=ql.Period(1, ql.Years),
                            fixedLegDayCount=ql.Thirty360(ql.Thirty360.BondBasis),
                            floatingLegSpread=0.0,
                            swapType=t)
                        underlying.setPricingEngine(self.swap_engine)

                        swaption = ql.Swaption(underlying,
                                               ql.EuropeanExercise(
                                                   exercise_date),
                                               settle_type,
                                               m)
                        swaption.setPricingEngine(swaption_engine)

                        annuity = swaption.annuity()
                        expected_annuity = 0.0
                        if (m == ql.Settlement.CollateralizedCashPrice):
                            expected_annuity = swap_pv01(underlying)
                        if (m == ql.Settlement.ParYieldCurve):
                            expected_annuity = par_yield_bps(
                                underlying, self.discount_handle)

                        fail_msg = """ Swaption annuity test failed for:
                                            option tenor: {option_tenor}
                                            volatility : {volatility}
                                            option type: {option_type}
                                            swap tenor: {swap_tenor}
                                            strike: {strike}
                                            settlement: {settle_type}
                                            method: {method}
                                            annuity: {annuity}
                                            replicated annuity: {expected_annuity}
                                   """.format(option_tenor=e,
                                              volatility=volatility,
                                              option_type=OPTION_TYPE_MAP[t],
                                              swap_tenor=l,
                                              strike=strike,
                                              settle_type=SETTLEMENT_TYPE_MAP[settle_type],
                                              method=SETTLEMENT_METHOD_MAP[m],
                                              annuity=annuity,
                                              expected_annuity=expected_annuity)
                        self.assertAlmostEqual(
                            first=annuity,
                            second=expected_annuity,
                            delta=EPSILON,
                            msg=fail_msg)

    def test_swaption_annuity_black_model(self):
        """Testing swaption annuity in Black model"""
        self._assert_swaption_annuity(
            swaption_pricer_func=make_const_black_vol_engine,
            use_bachelier_vol=False)

    def test_swaption_annuity_bachelier_model(self):
        """Testing swaption annuity in Bachelier model"""
        self._assert_swaption_annuity(
            swaption_pricer_func=make_const_bachelier_vol_engine,
            use_bachelier_vol=True)


if __name__ == "__main__":
    print("testing QuantLib", ql.__version__)
    unittest.main(verbosity=2)



# coding=utf-8-unix
"""
 Copyright (C) 2018 Wojciech Ślusarski

 This file is part of QuantLib, a free-software/open-source library
 for financial quantitative analysts and developers - http://quantlib.org/

 QuantLib is free software: you can redistribute it and/or modify it
 under the terms of the QuantLib license.  You should have received a
 copy of the license along with this program; if not, please email
 <quantlib-dev@lists.sf.net>. The license is also available online at
 <https://www.quantlib.org/license.shtml>.

 This program is distributed in the hope that it will be useful, but WITHOUT
 ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
 FOR A PARTICULAR PURPOSE.  See the license for more details.
"""

import QuantLib as ql
import unittest


class IborIndexTest(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.euribor3m = ql.Euribor3M()

    def setUp(self):
        self.euribor3m.clearFixings()
        # values are not real due to copyrights of the fixing
        self.euribor3m.addFixing(ql.Date(17, 7, 2018), -0.3)
        self.euribor3m.addFixings([ql.Date(12, 7, 2018), ql.Date(13, 7, 2018)], [-0.3, -0.3])

    def testAddFixingFail(self):
        """Testing for RuntimeError while trying to overwrite fixing value"""

        with self.assertRaises(RuntimeError):
            # attempt to overwrite value that is already set at different level
            self.euribor3m.addFixing(ql.Date(17, 7, 2018), -0.4)

        with self.assertRaises(RuntimeError):
            # attempt to overwrite value that is already set at different level
            self.euribor3m.addFixings([ql.Date(12, 7, 2018), ql.Date(13, 7, 2018)], [-0.4, -0.4])

    def testAddFixing(self):
        """Testing for overwriting fixing value"""

        force_overwrite = True
        try:
            # attempt to overwrite value that is already set at different level
            self.euribor3m.addFixing(ql.Date(17, 7, 2018), -0.4, force_overwrite)
            self.euribor3m.addFixings([ql.Date(12, 7, 2018), ql.Date(13, 7, 2018)], [-0.4, -0.4], force_overwrite)
            # try clearFixings and repeat with original levels
            self.euribor3m.clearFixings()
            self.euribor3m.addFixing(ql.Date(17, 7, 2018), -0.3)
            self.euribor3m.addFixings([ql.Date(12, 7, 2018), ql.Date(13, 7, 2018)], [-0.3, -0.3])

        except RuntimeError as err:
            raise AssertionError("Failed to overwrite index fixixng " + "{}".format(err))

    def testTimeSeries(self):
        """Testing for getting time series of the fixing"""

        dates = (ql.Date(12, 7, 2018), ql.Date(13, 7, 2018), ql.Date(17, 7, 2018))
        values = (-0.3, -0.3, -0.3)
        for expected, actual in zip(dates, self.euribor3m.timeSeries().dates()):
            self.assertTrue(expected == actual)
        for expected, actual in zip(values, self.euribor3m.timeSeries().values()):
            self.assertTrue(expected == actual)


class BMAIndexTest(unittest.TestCase):
    def setUp(self):
        self.index = ql.BMAIndex()

    def testInspectors(self):
        """Testing BMA index inspectors"""

        self.assertEqual(self.index.familyName(), "BMA")
        self.assertEqual(self.index.tenor(), ql.Period(1, ql.Weeks))
        self.assertEqual(self.index.fixingDays(), 1)
        self.assertEqual(self.index.currency(), ql.USDCurrency())

    def testFixingDates(self):
        """Testing BMA fixing-date logic"""

        self.assertTrue(self.index.isValidFixingDate(ql.Date(4, 1, 2023)))
        self.assertFalse(self.index.isValidFixingDate(ql.Date(5, 1, 2023)))

    def testMaturityDate(self):
        """Testing BMA maturity-date calculation"""

        self.assertEqual(
            self.index.maturityDate(ql.Date(5, 1, 2023)),
            ql.Date(12, 1, 2023))

    def testFixingSchedule(self):
        """Testing BMA fixing schedule"""

        schedule = self.index.fixingSchedule(
            ql.Date(5, 1, 2023), ql.Date(20, 1, 2023))
        self.assertEqual(schedule[0], ql.Date(4, 1, 2023))
        self.assertEqual(schedule[-1], ql.Date(25, 1, 2023))


if __name__ == "__main__":
    print("testing QuantLib", ql.__version__)
    unittest.main(verbosity=2)



# ---
# jupyter:
#   jupytext:
#     formats: py:light
#     text_representation:
#       extension: .py
#       format_name: light
#       format_version: '1.5'
#       jupytext_version: 1.4.2
#   kernelspec:
#     display_name: Python 3
#     language: python
#     name: python3
# ---

# # Credit default swaps
#
# Copyright (&copy;) 2014 Thema Consulting SA
#
# This file is part of QuantLib, a free-software/open-source library
# for financial quantitative analysts and developers - https://www.quantlib.org/
#
# QuantLib is free software: you can redistribute it and/or modify it under the
# terms of the QuantLib license.  You should have received a copy of the
# license along with this program; if not, please email
# <quantlib-dev@lists.sf.net>. The license is also available online at
# <https://www.quantlib.org/license.shtml>.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the license for more details.

# ### Setup

import QuantLib as ql

calendar = ql.TARGET()

todaysDate = ql.Date(15, ql.May, 2007)
ql.Settings.instance().evaluationDate = todaysDate

risk_free_rate = ql.YieldTermStructureHandle(ql.FlatForward(todaysDate, 0.01, ql.Actual365Fixed()))

# ### CDS parameters

recovery_rate = 0.5
quoted_spreads = [0.0150, 0.0150, 0.0150, 0.0150]
tenors = [ql.Period(3, ql.Months), ql.Period(6, ql.Months), ql.Period(1, ql.Years), ql.Period(2, ql.Years)]
maturities = [calendar.adjust(todaysDate + x, ql.Following) for x in tenors]

instruments = [
    ql.SpreadCdsHelper(
        ql.makeQuoteHandle(s),
        tenor,
        0,
        calendar,
        ql.Quarterly,
        ql.Following,
        ql.DateGeneration.TwentiethIMM,
        ql.Actual365Fixed(),
        recovery_rate,
        risk_free_rate,
    )
    for s, tenor in zip(quoted_spreads, tenors)
]

hazard_curve = ql.PiecewiseFlatHazardRate(todaysDate, instruments, ql.Actual365Fixed())
print("Calibrated hazard rate values: ")
for x in hazard_curve.nodes():
    print("hazard rate on %s is %.7f" % x)

print("Some survival probability values: ")
print(
    "1Y survival probability: %.4g, \n\t\texpected %.4g"
    % (hazard_curve.survivalProbability(todaysDate + ql.Period("1Y")), 0.9704)
)
print(
    "2Y survival probability: %.4g, \n\t\texpected %.4g"
    % (hazard_curve.survivalProbability(todaysDate + ql.Period("2Y")), 0.9418)
)

# ### Reprice instruments

nominal = 1000000.0
probability = ql.DefaultProbabilityTermStructureHandle(hazard_curve)

# We'll create a cds for every maturity:

all_cds = []
for maturity, s in zip(maturities, quoted_spreads):
    schedule = ql.Schedule(
        todaysDate,
        maturity,
        ql.Period(ql.Quarterly),
        calendar,
        ql.Following,
        ql.Unadjusted,
        ql.DateGeneration.TwentiethIMM,
        False,
    )
    cds = ql.CreditDefaultSwap(ql.Protection.Seller, nominal, s, schedule, ql.Following, ql.Actual365Fixed())
    engine = ql.MidPointCdsEngine(probability, recovery_rate, risk_free_rate)
    cds.setPricingEngine(engine)
    all_cds.append(cds)

print("Repricing of quoted CDSs employed for calibration: ")
for cds, tenor in zip(all_cds, tenors):
    print("%s fair spread: %.7g" % (tenor, cds.fairSpread()))
    print("   NPV: %g" % cds.NPV())
    print("   default leg: %.7g" % cds.defaultLegNPV())
    print("   coupon leg: %.7g" % cds.couponLegNPV())
    print("")
